import logging
from abc import ABC

import numpy as np
import pandas as pd
from tqdm import tqdm

from asreview.config import DEFAULT_N_INSTANCES
from asreview.config import LABEL_NA
from asreview.models.balance.simple import SimpleBalance
from asreview.models.classifiers import NaiveBayesClassifier
from asreview.models.feature_extraction.tfidf import Tfidf
from asreview.models.query.max import MaxQuery
from asreview.project import open_state
from asreview.settings import ASReviewSettings


class BaseReview(ABC):
    """Base class for Systematic Review.

    Arguments
    ---------
    as_data: asreview.ASReviewData
        The data object which contains the text, labels, etc.
    project: path-like
        Path to the project file.
    model: BaseTrainClassifier
        Initialized model to fit the data during active learning.
        See asreview.models.utils.py for possible models.
    query_model: BaseQueryStrategy
        Initialized model to query new instances for review, such as random
        sampling or max sampling.
        See asreview.query_strategies.utils.py for query models.
    balance_model: BaseBalance
        Initialized model to redistribute the training data during the
        active learning process. They might either resample or undersample
        specific papers.
    feature_model: BaseFeatureExtraction
        Feature extraction model that converts texts and keywords to
        feature matrices.
    n_instances: int
        Number of papers to query at each step in the active learning
        process.
    stop_if: int
        Number of steps/queries to perform. Set to None for no limit.
    start_idx: numpy.ndarray
        Start the simulation/review with these indices. They are assumed to
        be already labeled. Failing to do so might result bad behaviour.
    """

    def __init__(
        self,
        as_data,
        project,
        model=NaiveBayesClassifier(),
        query_model=MaxQuery(),
        balance_model=SimpleBalance(),
        feature_model=Tfidf(),
        n_papers=None,
        n_instances=DEFAULT_N_INSTANCES,
        stop_if=None,
        start_idx=[],
    ):
        """Initialize the reviewer base class, so that everything is ready to
        train a new model."""
        super(BaseReview, self).__init__()

        # Set the model.
        self.classifier = model
        self.balance_model = balance_model
        self.query_strategy = query_model
        self.feature_extraction = feature_model

        # Set the settings.
        self.as_data = as_data
        self.project = project
        self.n_instances = n_instances
        self.stop_if = stop_if
        self.prior_indices = start_idx

        if n_papers is not None:
            logging.warning("Argument n_papers is deprecated, ignoring n_papers.")

        # Get the known labels.
        self.data_labels = as_data.labels
        if self.data_labels is None:
            self.data_labels = np.full(len(as_data), LABEL_NA)

        with open_state(self.project, read_only=False) as state:
            # If the state is empty, add the settings.
            if state.is_empty():
                state.settings = self.settings

            # Add the record table to the state if it is not already there.
            self.record_table = state.get_record_table()
            if self.record_table.empty:
                state.add_record_table(as_data.record_ids)
                self.record_table = state.get_record_table()

            # Retrieve feature matrix from the project file or create
            # one from scratch.
            try:
                self.X = self.project.get_feature_matrix(self.feature_extraction.name)
            except FileNotFoundError:
                self.X = self.feature_extraction.fit_transform(
                    as_data.texts, as_data.headings, as_data.bodies, as_data.keywords
                )

                # check if the number of records after the transform equals
                # the number of records in the dataset
                if self.X.shape[0] != len(as_data):
                    raise ValueError(
                        "Dataset has {} records while feature "
                        "extractor returns {} records".format(
                            len(as_data), self.X.shape[0]
                        )
                    )

                self.project.add_feature_matrix(self.X, self.feature_extraction.name)

            # Check if the number or records in the feature matrix matches the
            # length of the dataset.
            if self.X.shape[0] != len(self.data_labels):
                raise ValueError(
                    "The state file does not correspond to the "
                    "given data file, please use another state "
                    "file or dataset."
                )

            # Make sure the priors are labeled.
            self._label_priors()

    @property
    def settings(self):
        """Get an ASReview settings object"""
        extra_kwargs = {}
        if hasattr(self, "n_prior_included"):
            extra_kwargs["n_prior_included"] = self.n_prior_included
        if hasattr(self, "n_prior_excluded"):
            extra_kwargs["n_prior_excluded"] = self.n_prior_excluded
        return ASReviewSettings(
            model=self.classifier.name,
            query_strategy=self.query_strategy.name,
            balance_strategy=self.balance_model.name,
            feature_extraction=self.feature_extraction.name,
            n_instances=self.n_instances,
            stop_if=self.stop_if,
            model_param=self.classifier.param,
            query_param=self.query_strategy.param,
            balance_param=self.balance_model.param,
            feature_param=self.feature_extraction.param,
            **extra_kwargs
        )

    def review(self):
        """Do a full review."""
        # Label any pending records.

        with open_state(self.project, read_only=False) as s:
            pending = s.get_pending()
            if not pending.empty:
                self._label(pending)

            labels_prior = s.get_labels()

        # progress bars
        pbar_rel = tqdm(
            initial=sum(labels_prior),
            total=sum(self.as_data.labels),
            desc="Relevant records found",
        )
        pbar_total = tqdm(
            initial=len(labels_prior),
            total=len(self.as_data),
            desc="Records labeled       ",
        )

        # While the stopping condition has not been met:
        while not self._stop_review():
            # Train a new model.
            self.train()

            # Query for new records to label.
            record_ids = self._query(self.n_instances)

            # Label the records.
            labels = self._label(record_ids)

            # monitor progress here
            pbar_rel.update(sum(labels))
            pbar_total.update(len(labels))

        else:
            # write to state when stopped
            pbar_rel.close()
            pbar_total.close()

    def _label_priors(self):
        """Make sure the prior records are labeled."""
        with open_state(self.project, read_only=False) as state:
            labeled = state.get_labeled()
            unlabeled_priors = [
                x for x in self.prior_indices if x not in labeled["record_id"].to_list()
            ]
            self._label(unlabeled_priors, prior=True)

    def _stop_review(self):
        """Check if the review should be stopped according to stopping rule
        obtained from the settings.

        Returns
        -------
        bool
            If True, the stopping criteria have been met.
        """
        stop = False

        # Get the pool and labeled. There never should be pending papers here.
        with open_state(self.project) as state:
            pool, labeled, _ = state.get_pool_labeled_pending()

        # if the pool is empty, always stop
        if pool.empty:
            stop = True

        # If stop_if is set to min, stop when all papers in the pool are
        # irrelevant.
        if self.stop_if == "min" and (self.data_labels[pool] == 0).all():
            stop = True
        # Otherwise, stop when reaching stop_if (if provided)
        elif self.stop_if is not None:
            with open_state(self.project) as state:
                training_sets = state.get_training_sets()
                # There is one query per trained model. We subtract 1
                # for the priors.
                stop_if = len(set(training_sets)) - 1
            if stop_if >= self.stop_if:
                stop = True

        return stop

    def _query(self, n):
        """Query new records to label.

        Arguments
        ---------
        n: int
            Number of records to query.

        Returns
        -------
        list
            List of record_ids of the n top ranked records according to the last
            ranking saved in the state.
        """
        with open_state(self.project, read_only=False) as s:
            top_n_records = s.query_top_ranked(n)
        return top_n_records

    def _label(self, record_ids, prior=False):
        """Label queried records uses the known labels in a simulated review.

        Arguments
        ---------
        record_ids: list
            List of record_ids that will be labeled.
        prior: bool
            Whether the records priors or not.
        """
        labels = self.data_labels[record_ids]

        with open_state(self.project, read_only=False) as s:
            s.add_labeling_data(record_ids, labels, prior=prior)

    def train(self):
        """Train a new model on the labeled data."""
        # Check if both labels are available.
        with open_state(self.project) as state:
            labeled = state.get_labeled()
            labels = labeled["label"].to_list()
            training_set = len(labeled)
            if not (0 in labels and 1 in labels):
                raise ValueError(
                    "Not both labels available. " "Stopped training the model"
                )

        # TODO: Simplify balance model input.
        # Use the balance model to sample the trainings data.
        y_sample_input = (
            pd.DataFrame(self.record_table)
            .merge(labeled, how="left", on="record_id")
            .loc[:, "label"]
            .fillna(LABEL_NA)
            .to_numpy()
        )
        train_idx = np.where(y_sample_input != LABEL_NA)[0]

        X_train, y_train = self.balance_model.sample(self.X, y_sample_input, train_idx)

        # Fit the classifier on the trainings data.
        self.classifier.fit(X_train, y_train)

        # Use the query strategy to produce a ranking.
        ranked_record_ids = self.query_strategy.query(
            self.X, classifier=self.classifier
        )

        # TODO: Also log the probablities.
        # Log the ranking in the state.
        with open_state(self.project, read_only=False) as state:
            state.add_last_ranking(
                ranked_record_ids,
                self.classifier.name,
                self.query_strategy.name,
                self.balance_model.name,
                self.feature_extraction.name,
                training_set,
            )



# Copyright 2019-2022 The ASReview Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np


def n_records(data):
    """Return the number of records.

    Arguments
    ---------
    data: asreview.data.ASReviewData
        An ASReviewData object with the records.

    Return
    ------
    int:
        The statistic
    """
    return len(data)


def n_relevant(data):
    """Return the number of relevant records.

    Arguments
    ---------
    data: asreview.data.ASReviewData
        An ASReviewData object with the records.

    Return
    ------
    int:
        The statistic
    """
    if data.labels is not None:
        return len(np.where(data.labels == 1)[0])
    return None


def n_irrelevant(data):
    """Return the number of irrelevant records.

    Arguments
    ---------
    data: asreview.data.ASReviewData
        An ASReviewData object with the records.

    Return
    ------
    int:
        The statistic
    """
    if data.labels is None:
        return None
    return len(np.where(data.labels == 0)[0])


def n_unlabeled(data):
    """Return the number of unlabeled records.

    Arguments
    ---------
    data: asreview.data.ASReviewData
        An ASReviewData object with the records.

    Return
    ------
    int:
        The statistic
    """
    if data.labels is None:
        return None
    return len(data.labels) - n_relevant(data) - n_irrelevant(data)


def n_missing_title(data):
    """Return the number of records with missing titles.

    Arguments
    ---------
    data: asreview.data.ASReviewData
        An ASReviewData object with the records.

    Return
    ------
    int:
        The statistic
    """
    n_missing = 0
    if data.title is None:
        return None, None
    if data.labels is None:
        n_missing_included = None
    else:
        n_missing_included = 0
    for i in range(len(data.title)):
        if len(data.title[i]) == 0:
            n_missing += 1
            if data.labels is not None and data.labels[i] == 1:
                n_missing_included += 1
    return n_missing, n_missing_included


def n_missing_abstract(data):
    """Return the number of records with missing abstracts.

    Arguments
    ---------
    data: asreview.data.ASReviewData
        An ASReviewData object with the records.

    Return
    ------
    int:
        The statistic
    """
    n_missing = 0
    if data.abstract is None:
        return None, None
    if data.labels is None:
        n_missing_included = None
    else:
        n_missing_included = 0

    for i in range(len(data.abstract)):
        if len(data.abstract[i]) == 0:
            n_missing += 1
            if data.labels is not None and data.labels[i] == 1:
                n_missing_included += 1

    return n_missing, n_missing_included


def title_length(data):
    """Return the average length of the titles.

    Arguments
    ---------
    data: asreview.data.ASReviewData
        An ASReviewData object with the records.

    Return
    ------
    int:
        The statistic
    """
    if data.title is None:
        return None
    avg_len = 0
    for i in range(len(data.title)):
        avg_len += len(data.title[i])
    return avg_len / len(data.title)


def abstract_length(data):
    """Return the average length of the abstracts.

    Arguments
    ---------
    data: asreview.data.ASReviewData
        An ASReviewData object with the records.

    Return
    ------
    int:
        The statistic
    """
    if data.abstract is None:
        return None
    avg_len = 0
    for i in range(len(data.abstract)):
        avg_len += len(data.abstract[i])
    return avg_len / len(data.abstract)


def n_keywords(data):
    """Return the number of keywords.

    Arguments
    ---------
    data: asreview.data.ASReviewData
        An ASReviewData object with the records.

    Return
    ------
    int:
        The statistic
    """
    if data.keywords is None:
        return None
    return np.average([len(keywords) for keywords in data.keywords])


def n_duplicates(data, pid="doi"):
    """Number of duplicates.

    Duplicate detection can be a very challenging task. Multiple
    algorithms can be used and results can be vary.

    Arguments
    ---------
    data: asreview.data.ASReviewData
        An ASReviewData object with the records.
    pid: string
        Which persistent identifier (PID) to use for deduplication.
        Default is 'doi'.

    Return
    ------
    int:
        Number of duplicates
    """
    return int(data.duplicated(pid).sum())



from pathlib import Path

import pytest
from pandas.testing import assert_frame_equal

from asreview.entry_points.simulate import SimulateEntryPoint
from asreview.project import open_state

DATA_FP = Path("tests", "demo_data", "generic_labels.csv")


@pytest.mark.parametrize(
    "seed",
    [
        (535),
        (165),
        (42),
    ],
)
def test_init_seed(tmpdir, seed):
    project1_fp = Path(tmpdir, "tmp_state1.asreview")
    project2_fp = Path(tmpdir, "tmp_state2.asreview")

    entry_point = SimulateEntryPoint()

    # simulate run 1
    argv1 = (
        f"{DATA_FP} -s {project1_fp} -m nb --init_seed"
        f" {seed} --n_prior_excluded 1 --n_prior_included 1 -n 2".split()
    )

    # simulate run 2
    argv2 = (
        f"{DATA_FP} -s {project2_fp} -m nb --init_seed"
        f" {seed} --n_prior_excluded 1 --n_prior_included 1  -n 2".split()
    )

    # run the simulations
    entry_point.execute(argv1)
    entry_point.execute(argv2)

    # open the state file and extract the priors
    with open_state(project1_fp) as s1:
        record_ids1 = s1.get_priors()["record_id"]

    with open_state(project2_fp) as s2:
        record_ids2 = s2.get_priors()["record_id"]

    assert record_ids1.tolist() == record_ids2.tolist()


def test_no_seed(tmpdir):
    priors = []
    for i in range(20):
        # get project url
        project_fp = Path(tmpdir, f"tmp_state_{i}.asreview")

        entry_point = SimulateEntryPoint()
        argv = (
            f"{DATA_FP} -s {project_fp} -m nb "
            f"--n_prior_excluded 1 --n_prior_included 1 -n 2".split()
        )
        entry_point.execute(argv)

        # open the state file and extract the priors
        with open_state(project_fp) as s:
            priors.extend(s.get_priors()["record_id"].tolist())

    assert len(set(priors)) > 2


@pytest.mark.parametrize(
    "seed",
    [
        (535),
        (165),
        (42),
    ],
)
def test_model_seed(tmpdir, seed):
    project1_fp = Path(tmpdir, "tmp_state1.asreview")
    project2_fp = Path(tmpdir, "tmp_state2.asreview")

    entry_point = SimulateEntryPoint()

    # simulate run 1
    argv1 = (
        f"{DATA_FP} -s {project1_fp} -m rf --init_seed {seed}"
        f" --seed {seed}"
        f" --n_prior_excluded 1 --n_prior_included 1".split()
    )

    # simulate run 2
    argv2 = (
        f"{DATA_FP} -s {project2_fp} -m rf --init_seed {seed}"
        f" --seed {seed}"
        f" --n_prior_excluded 1 --n_prior_included 1".split()
    )

    # run the simulations
    entry_point.execute(argv1)
    entry_point.execute(argv2)

    # open the state file and extract the priors
    with open_state(project1_fp) as s1:
        record_table1 = s1.get_dataset().drop("labeling_time", axis=1)

    with open_state(project2_fp) as s2:
        record_table2 = s2.get_dataset().drop("labeling_time", axis=1)

    assert_frame_equal(record_table1, record_table2)



# Copyright 2019-2022 The ASReview Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from math import ceil

import numpy as np

from asreview.models.balance.base import BaseBalance
from asreview.utils import get_random_state


class UndersampleBalance(BaseBalance):
    """Undersampling balance strategy (``undersample``).

    This undersamples the data, leaving out excluded papers so that the
    included and excluded papers are in some particular ratio (closer to one).

    Arguments
    ---------
    ratio: double
        Undersampling ratio of the zero's. If for example we set a ratio of
        0.25, we would sample only a quarter of the zeros and all the ones.
    """

    name = "undersample"
    label = "Undersampling"

    def __init__(self, ratio=1.0, random_state=None):
        """Initialize the undersampling balance strategy."""
        super(UndersampleBalance, self).__init__()
        self.ratio = ratio
        self._random_state = get_random_state(random_state)

    def sample(self, X, y, train_idx):
        """Resample the training data.

        Arguments
        ---------
        X: numpy.ndarray
            Complete feature matrix.
        y: numpy.ndarray
            Labels for all papers.
        train_idx: numpy.ndarray
            Training indices, that is all papers that have been reviewed.

        Returns
        -------
        numpy.ndarray,numpy.ndarray:
            X_train, y_train: the resampled matrix, labels.
        """
        one_ind = train_idx[np.where(y[train_idx] == 1)]
        zero_ind = train_idx[np.where(y[train_idx] == 0)]

        n_one = len(one_ind)
        n_zero = len(zero_ind)

        # If we don't have an excess of 0's, give back all training_samples.
        if n_one / n_zero >= self.ratio:
            shuf_ind = np.append(one_ind, zero_ind)
        else:
            n_zero_epoch = ceil(n_one / self.ratio)
            zero_under = self._random_state.choice(
                np.arange(n_zero), n_zero_epoch, replace=False
            )
            shuf_ind = np.append(one_ind, zero_ind[zero_under])

        self._random_state.shuffle(shuf_ind)
        return X[shuf_ind], y[shuf_ind]

    def full_hyper_space(self):
        from hyperopt import hp

        parameter_space = {
            "bal_ratio": hp.lognormal("bal_ratio", 0, 1),
        }
        return parameter_space, {}



# Copyright 2019-2022 The ASReview Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from functools import wraps

from flask import current_app
from flask import jsonify
from flask import request
from flask_login import current_user
from flask_login.config import EXEMPT_METHODS

from asreview.webapp.authentication.models import Project


def asreview_login_required(func):
    @wraps(func)
    def decorated_view(*args, **kwargs):
        if not current_app.config.get("AUTHENTICATION_ENABLED"):
            pass
        elif request.method in EXEMPT_METHODS:
            pass
        else:
            if not (bool(current_user) and current_user.is_authenticated):
                return jsonify({"message": "Login required."}), 401

        return func(*args, **kwargs)

    return decorated_view


def project_authorization(f):
    """Decorator function that checks if current user can access
    a project in an authenticated situation"""

    @wraps(f)
    def decorated_function(project_id, *args, **kwargs):
        if app_is_authenticated(current_app):
            # find the project
            project = Project.query.filter(
                Project.project_id == project_id
            ).one_or_none()
            if project is None:
                return jsonify({"message": "project not found"}), 404
            # if there is a project, check if
            all_users = set([project.owner] + project.collaborators)
            if current_user not in all_users:
                return jsonify({"message": "no permission"}), 403
        return f(project_id, *args, **kwargs)

    return decorated_function


def app_is_authenticated(app):
    """Checks is app is authenticated"""
    return app.config.get("AUTHENTICATION_ENABLED", False)



# Copyright 2019-2022 The ASReview Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import os
from configparser import ConfigParser

from asreview.config import DEFAULT_N_INSTANCES
from asreview.models.balance import get_balance_model
from asreview.models.classifiers import get_classifier
from asreview.models.feature_extraction import get_feature_model
from asreview.models.query import get_query_model
from asreview.types import type_n_queries
from asreview.utils import pretty_format

SETTINGS_TYPE_DICT = {
    "model": str,
    "query_strategy": str,
    "balance_strategy": str,
    "feature_extraction": str,
    "n_instances": int,
    "stop_if": type_n_queries,
    "n_prior_included": int,
    "n_prior_excluded": int,
    "mode": str,
    "model_param": dict,
    "query_param": dict,
    "feature_param": dict,
    "balance_param": dict,
}


def _map_settings_type(name, value):
    if value is None:
        return None

    try:
        return SETTINGS_TYPE_DICT[name](value)
    except TypeError:
        raise TypeError(f"Can't convert setting '{name}' to {SETTINGS_TYPE_DICT[name]}")


def _convert_types(par_defaults, param):
    """Convert strings from the config file to the appropriate type."""
    for par in param:
        try:
            par_type = type(par_defaults[par])
            if par_type == bool:
                param[par] = param[par] in ["True", "true", "T", "t", True]
            else:
                try:
                    param[par] = par_type(param[par])
                except TypeError:
                    raise TypeError(f"Error converting key in config file: {par}")
        except KeyError:
            logging.warning(
                f"Parameter {par} does not have a default.\n"
                f"Defaults: {par_defaults}."
            )


class ASReviewSettings(object):
    """Object to store the configuration of a review session.

    The main difference being that it type checks (some)
    of its contents.
    """

    def __init__(
        self,
        model,
        query_strategy,
        balance_strategy,
        feature_extraction,
        n_instances=DEFAULT_N_INSTANCES,
        stop_if=None,
        n_prior_included=None,
        n_prior_excluded=None,
        as_data=None,
        model_param={},
        query_param={},
        balance_param={},
        feature_param={},
        data_fp=None,
        n_queries=None,
        abstract_only=False,  # deprecated
        mode=None,  # deprecated
        n_papers=None,  # deprecated
        data_name=None,  # deprecated
    ):
        self.model = model
        self.query_strategy = query_strategy
        self.balance_strategy = balance_strategy
        self.feature_extraction = feature_extraction
        self.n_instances = n_instances
        self.stop_if = stop_if
        self.n_prior_included = n_prior_included
        self.n_prior_excluded = n_prior_excluded
        self.as_data = as_data
        self.model_param = model_param
        if query_strategy == "max_random":
            query_param_copy = query_param.copy()
            try:
                del query_param_copy["strategy_1"]
                del query_param_copy["strategy_2"]
            except KeyError:
                pass
            self.query_param = query_param_copy
        else:
            self.query_param = query_param
        self.balance_param = balance_param
        self.feature_param = feature_param

    def __str__(self):
        return pretty_format(self.to_dict())

    def __setattr__(self, name, value):
        try:
            super(ASReviewSettings, self).__setattr__(
                name, _map_settings_type(name, value)
            )
        except KeyError:
            super(ASReviewSettings, self).__setattr__(name, value)

    def to_dict(self):
        """Export default settings to dict."""
        info_dict = {}
        for attrib in SETTINGS_TYPE_DICT:
            value = getattr(self, attrib, None)
            if value is not None:
                info_dict[attrib] = value
        return info_dict

    def from_file(self, config_file):
        """Fill the contents of settings by reading a config file.

        Arguments
        ---------
        config_file: str
            Source configuration file.

        """
        if config_file is None or not os.path.isfile(config_file):
            if config_file is not None:
                print(f"Didn't find configuration file: {config_file}")
            return

        config = ConfigParser()
        config.optionxform = str
        config.read(config_file)

        # Read the each of the sections.
        for sect in config:
            if sect == "global_settings":
                for key, value in config.items(sect):
                    try:
                        setattr(self, key, SETTINGS_TYPE_DICT[key](value))
                    except (KeyError, TypeError):
                        print(
                            f"Warning: value with key '{key}' is ignored "
                            "(spelling mistake, wrong type?)."
                        )

            elif sect in [
                "model_param",
                "query_param",
                "balance_param",
                "feature_param",
            ]:
                setattr(self, sect, dict(config.items(sect)))
            elif sect != "DEFAULT":
                print(
                    f"Warning: section [{sect}] is ignored in "
                    f"config file {config_file}"
                )

        model = get_classifier(self.model)
        _convert_types(model.default_param, self.model_param)
        balance_model = get_balance_model(self.balance_strategy)
        _convert_types(balance_model.default_param, self.balance_param)
        query_model = get_query_model(self.query_strategy)
        _convert_types(query_model.default_param, self.query_param)
        feature_model = get_feature_model(self.feature_extraction)
        _convert_types(feature_model.default_param, self.feature_param)



# Copyright 2019-2022 The ASReview Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import sqlite3
from datetime import datetime
from pathlib import Path

import pandas as pd

from asreview._version import get_versions
from asreview.settings import ASReviewSettings
from asreview.state.base import BaseState
from asreview.state.errors import StateError
from asreview.state.errors import StateNotFoundError

REQUIRED_TABLES = [
    # the table with the labeling decisions and models trained
    "results",
    # the mapping of record identifiers to row numbers
    "record_table",
    # the latest probabilities.
    "last_probabilities",
    # the latest ranking.
    "last_ranking",
    # the record ids whose labeling decision was changed.
    "decision_changes",
]

RESULTS_TABLE_COLUMNS = [
    "record_id",
    "label",
    "classifier",
    "query_strategy",
    "balance_strategy",
    "feature_extraction",
    "training_set",
    "labeling_time",
    "notes",
]
SETTINGS_METADATA_KEYS = [
    "settings",
    "state_version",
    "software_version",
    "model_has_trained",
]


class SQLiteState(BaseState):
    """Class for storing the review state.

    The results are stored in a sqlite database.

    Arguments
    ---------
    read_only: bool
        Open state in read only mode. Default False.

    Attributes
    ----------
    version: str
        Return the version number of the state.
    settings: asreview.settings.ASReviewSettings
        Return an ASReview settings object with model settings and
        active learning settings.
    n_records_labeled: int
        Get the number of labeled records, where each prior is counted
        individually.
    n_priors: int
        Number of priors. If priors have not been selected returns None.
    exist_new_labeled_records: bool
        Have there been labeled records added to the state since the last time
        a model ranking was added to the state?
    model_has_trained: bool
        Has the ranking by a model been added to the state?
    """

    def __init__(self, read_only=True):
        super(SQLiteState, self).__init__(read_only=read_only)

    # INTERNAL PATHS AND CONNECTIONS

    def _connect_to_sql(self):
        """Get a connection to the SQLite database.

        Returns
        -------
        sqlite3.Connection
            Connection to the the SQLite database.
            The connection is read only if self.read_only is true.
        """
        if self.read_only:
            con = sqlite3.connect(f"file:{str(self._sql_fp)}?mode=ro", uri=True)
        else:
            con = sqlite3.connect(str(self._sql_fp))
        return con

    @property
    def _sql_fp(self):
        """Get the path to the sqlite database."""

        return Path(self.review_dir, "results.sql")

    @property
    def _settings_metadata_fp(self):
        """Get the path to the settings and metadata json file."""

        return Path(self.review_dir, "settings_metadata.json")

    def _create_new_state_file(self, working_dir, review_id):
        """Create the files for storing a new state given an review_id.

        Stages:
        1: create result structure
        2: create model settings
        3: add state to the project file

        Arguments
        ---------
        review_dir: str, pathlib.Path
            Review folder location.
        review_id: str
            Identifier of the review.
        """
        if self.read_only:
            raise ValueError("Can't create new state file in read_only mode.")

        self.review_dir = Path(working_dir, "reviews", review_id)

        # create folder in the folder `results` with the name of result_id
        self._sql_fp.parent.mkdir(parents=True, exist_ok=True)

        # Create results table.
        con = self._connect_to_sql()
        try:
            cur = con.cursor()

            # Create the results table.
            cur.execute(
                """CREATE TABLE results
                                (record_id INTEGER,
                                label INTEGER,
                                classifier TEXT,
                                query_strategy TEXT,
                                balance_strategy TEXT,
                                feature_extraction TEXT,
                                training_set INTEGER,
                                labeling_time INTEGER,
                                notes TEXT)"""
            )

            # Create the record table.
            cur.execute(
                """CREATE TABLE record_table
                                (record_id INT)"""
            )

            # Create the last_probabilities table.
            cur.execute(
                """CREATE TABLE last_probabilities
                                (proba REAL)"""
            )

            # Create the last_ranking table.
            cur.execute(
                """CREATE TABLE last_ranking
                                (record_id INTEGER,
                                ranking INT,
                                classifier TEXT,
                                query_strategy TEXT,
                                balance_strategy TEXT,
                                feature_extraction TEXT,
                                training_set INTEGER,
                                time INTEGER)"""
            )

            # Create the table of changed decisions.
            cur.execute(
                """CREATE TABLE decision_changes
                                (record_id INTEGER,
                                new_label INTEGER,
                                time INTEGER)"""
            )

            con.commit()
            con.close()
        except sqlite3.Error as e:
            con.close()
            raise e

        # Create settings_metadata.json file
        # content of the settings is added later
        self.settings_metadata = {
            "settings": None,
            "state_version": "1",
            "software_version": get_versions()["version"],
            "model_has_trained": False,
        }

        with open(self._settings_metadata_fp, "w") as f:
            json.dump(self.settings_metadata, f)

    def _restore(self, working_dir, review_id):
        """
        Restore a state from files.

        Arguments
        ---------
        review_dir: str, pathlib.Path
            Review folder location.
        review_id: str
            Identifier of the review.
        """
        # store filepath
        self.review_dir = Path(working_dir, "reviews", review_id)

        # If state already exist
        if not working_dir.is_dir():
            raise StateNotFoundError(f"Project {working_dir} doesn't exist.")

        if not self._sql_fp.parent.is_dir():
            raise StateNotFoundError(f"Review with id {review_id} doesn't exist.")

        # Cache the settings.
        try:
            with open(self._settings_metadata_fp, "r") as f:
                self.settings_metadata = json.load(f)
        except FileNotFoundError:
            raise AttributeError(
                "'settings_metadata.json' not found in the state file."
            )

        try:
            if not self._is_valid_version():
                raise ValueError(
                    f"State cannot be read: state version {self.version}, "
                    f"state file version {self.version}."
                )
        except AttributeError as err:
            raise ValueError(f"Unexpected error when opening state file: {err}")

        self._is_valid_state()

    def _is_valid_state(self):
        con = self._connect_to_sql()
        cur = con.cursor()
        column_names = cur.execute("PRAGMA table_info(results)").fetchall()
        table_names = cur.execute(
            "SELECT name FROM sqlite_master WHERE type='table';"
        ).fetchall()
        con.close()

        # Check if all required tables are present.
        table_names = [tup[0] for tup in table_names]
        missing_tables = [
            table for table in REQUIRED_TABLES if table not in table_names
        ]
        if missing_tables:
            raise StateError(
                f"The SQL file should contain tables named "
                f"'{' '.join(missing_tables)}'."
            )

        # Check if all required columns are present in results.
        column_names = [tup[1] for tup in column_names]
        missing_columns = [
            col for col in RESULTS_TABLE_COLUMNS if col not in column_names
        ]
        if missing_columns:
            raise StateError(
                f"The results table does not contain the columns "
                f"{' '.join(missing_columns)}."
            )

        # Check settings_metadata contains the required keys.
        missing_keys = [
            key
            for key in SETTINGS_METADATA_KEYS
            if key not in self.settings_metadata.keys()
        ]
        if missing_keys:
            raise StateError(
                f"The keys {' '.join(missing_keys)} were not found in "
                f"settings_metadata."
            )

    def close(self):
        pass

    # PROPERTIES

    def _is_valid_version(self):
        """Check compatibility of state version."""
        return self.version[0] == "1"

    @property
    def version(self):
        """Version number of the state.

        Returns
        -------
        str:
            Returns the version of the state.

        """
        try:
            return self.settings_metadata["state_version"]
        except KeyError:
            raise AttributeError(
                "'settings_metadata.json' does not contain 'state_version'."
            )

    @property
    def settings(self):
        """Settings of the ASReview pipeline.

        Example
        -------

        Example of settings.

            model             : nb
            query_strategy    : max_random
            balance_strategy  : triple
            feature_extraction: tfidf
            n_instances       : 1
            stop_if           : min
            n_prior_included  : 10
            n_prior_excluded  : 10
            mode              : simulate
            model_param       : {'alpha': 3.822}
            query_param       : {'strategy_1': 'max', 'strategy_2': 'random',
            'mix_ratio': 0.95}
            feature_param     : {}
            balance_param     : {'a': 2.155, 'alpha': 0.94, ... 'gamma': 2.0,
            'shuffle': True}
            abstract_only     : False

        """
        settings = self.settings_metadata["settings"]
        if settings is None:
            return None
        return ASReviewSettings(**settings)

    @settings.setter
    def settings(self, settings):
        if isinstance(settings, ASReviewSettings):
            self._add_settings_metadata("settings", settings.to_dict())
        else:
            raise ValueError("'settings' should be an ASReviewSettings object.")

    @property
    def n_records(self):
        """Number of records in the loop.

        Returns
        -------
        int
            Number of records.
        """
        con = self._connect_to_sql()
        cur = con.cursor()
        cur.execute("SELECT COUNT (*) FROM record_table")
        n = cur.fetchone()[0]
        con.close()

        return n

    @property
    def n_records_labeled(self):
        """Number labeled records.

        Returns
        -------
        int
            Number of labeled records, priors counted individually.
        """
        labeled = self.get_labeled()
        return len(labeled)

    @property
    def n_priors(self):
        """Number of records added as prior knowledge.

        Returns
        -------
        int
            Number of records which were added as prior knowledge.
        """
        con = self._connect_to_sql()
        cur = con.cursor()
        cur.execute("SELECT COUNT (*) FROM results WHERE query_strategy='prior'")
        n = cur.fetchone()
        con.close()
        n = n[0]

        if n == 0:
            return None
        return n

    @property
    def exist_new_labeled_records(self):
        """Return True if there are new labeled records.

        Return True if there are any record labels added since the last time
        the model ranking was added to the state. Also returns True if no
        model was trained yet, but priors have been added.
        """
        labeled = self.get_labeled()
        last_training_set = self.get_last_ranking()["training_set"]
        if last_training_set.empty:
            return len(labeled) > 0
        else:
            return len(labeled) > last_training_set.iloc[0]

    @property
    def model_has_trained(self):
        """Return True if there is data of a trained model in the state."""
        return self.settings_metadata["model_has_trained"]

    def _add_settings_metadata(self, key, value):
        """Add information to the settings_metadata dictionary."""
        if self.read_only:
            raise ValueError("Can't change settings in read only mode.")
        self.settings_metadata[key] = value

        with open(self._settings_metadata_fp, "w") as f:
            json.dump(self.settings_metadata, f)

    def add_record_table(self, record_ids):
        """Add the record table to the state.

        Arguments
        ---------
        record_ids: list, np.array
            List containing all record ids of the dataset.
        """
        record_sql_input = [(int(record_id),) for record_id in record_ids]

        con = self._connect_to_sql()
        cur = con.cursor()
        cur.execute("DELETE FROM record_table")
        cur.executemany(
            "INSERT INTO record_table (record_id) VALUES (?)", record_sql_input
        )
        con.commit()

    def add_last_probabilities(self, probabilities):
        """Save the probabilities produced by the last classifier.

        Arguments
        ---------
        probabilities: list, np.array
            List containing the probabilities for every record.
        """
        proba_sql_input = [(proba,) for proba in probabilities]

        con = self._connect_to_sql()
        cur = con.cursor()

        # Check that the number of rows in the table is 0 (if the table is not
        # yet populated), or that it's equal to len(probabilities).
        cur.execute("SELECT COUNT (*) FROM last_probabilities")
        proba_length = cur.fetchone()[0]
        if not ((proba_length == 0) or (proba_length == len(proba_sql_input))):
            raise ValueError(
                f"There are {proba_length} probabilities in the database, "
                f"but 'probabilities' has length {len(probabilities)}"
            )

        cur.execute("""DELETE FROM last_probabilities""")
        cur.executemany(
            "INSERT INTO last_probabilities (proba) VALUES (?)", proba_sql_input
        )
        con.commit()

    def add_last_ranking(
        self,
        ranked_record_ids,
        classifier,
        query_strategy,
        balance_strategy,
        feature_extraction,
        training_set,
    ):
        """Save the ranking of the last iteration of the model.

        Save the ranking of the last iteration of the model, in the ranking
        order, so the record on row 0 is ranked first by the model.

        Arguments
        ---------
        ranked_record_ids: list, numpy.ndarray
            A list of records ids in the order that they were ranked.
        classifier: str
            Name of the classifier of the model.
        query_strategy: str
            Name of the query strategy of the model.
        balance_strategy: str
            Name of the balance strategy of the model.
        feature_extraction: str
            Name of the feature extraction method of the model.
        training_set: int
            Number of labeled records available at the time of training.
        """
        record_ids = self.get_record_table()

        if len(record_ids) != len(ranked_record_ids):
            raise ValueError(
                "The ranking should have the same length as the " "record table."
            )

        ranking = range(len(record_ids))
        classifiers = [classifier for _ in record_ids]
        query_strategies = [query_strategy for _ in record_ids]
        balance_strategies = [balance_strategy for _ in record_ids]
        feature_extractions = [feature_extraction for _ in record_ids]
        training_sets = [int(training_set) for _ in record_ids]
        ranking_times = [datetime.now()] * len(record_ids)

        # Create the database rows.
        db_rows = [
            (
                int(ranked_record_ids[i]),
                int(ranking[i]),
                classifiers[i],
                query_strategies[i],
                balance_strategies[i],
                feature_extractions[i],
                training_sets[i],
                ranking_times[i],
            )
            for i in range(len(record_ids))
        ]

        con = self._connect_to_sql()
        cur = con.cursor()
        cur.execute("DELETE FROM last_ranking")
        cur.executemany(
            (
                "INSERT INTO last_ranking (record_id, ranking, classifier, "
                "query_strategy, balance_strategy, feature_extraction, "
                "training_set, time) VALUES (?, ?, ?, ?, ?, ?, ?, ?)"
            ),
            db_rows,
        )
        con.commit()
        con.close()

        # If it's the first ranking table to be added, set model_has_trained.
        if not self.model_has_trained:
            self._add_settings_metadata("model_has_trained", True)

    def add_note(self, note, record_id):
        """Add a text note to save with a labeled record.

        Arguments
        ---------
        note: str
            Text note to save.
        record_id: int
            Identifier of the record to which the note should be added.
        """
        con = self._connect_to_sql()
        cur = con.cursor()
        cur.execute(
            "UPDATE results SET notes = ? WHERE record_id = ?", (note, record_id)
        )
        con.commit()
        con.close()

    def add_labeling_data(self, record_ids, labels, notes=None, prior=False):
        """Add the data corresponding to a labeling action to the state file.

        Arguments
        ---------
        record_ids: list, numpy.ndarray
            A list of ids of the labeled records as int.
        labels: list, numpy.ndarray
            A list of labels of the labeled records as int.
        notes: list of str/None
            A list of text notes to save with the labeled records.
        prior: bool
            Whether the added record are prior knowledge.
        """

        # Check if the state is still valid.
        self._is_valid_state()

        labeling_times = [datetime.now()] * len(record_ids)

        if notes is None:
            notes = [None for _ in record_ids]

        lengths = [len(record_ids), len(labels), len(notes)]
        # Check that all input data has the same length.
        if len(set(lengths)) != 1:
            raise ValueError("Input data should be of the same length.")
        n_records_labeled = len(record_ids)

        pool, _, pending = self.get_pool_labeled_pending()

        if prior:
            # Check that the record_ids are in the pool.
            if not all(record_id in pool.values for record_id in record_ids):
                raise ValueError(
                    "Labeling priors, but not all " "record_ids were found in the pool."
                )

            query_strategies = ["prior" for _ in record_ids]
            training_sets = [-1 for _ in record_ids]
            data = [
                (
                    int(record_ids[i]),
                    int(labels[i]),
                    query_strategies[i],
                    training_sets[i],
                    labeling_times[i],
                    notes[i],
                )
                for i in range(n_records_labeled)
            ]

            # If prior, we need to insert new records into the database.
            query = (
                "INSERT INTO results (record_id, label, query_strategy, "
                "training_set, labeling_time, notes) "
                "VALUES (?, ?, ?, ?, ?, ?)"
            )

        else:
            # Check that the record_ids are pending.
            if not all(record_id in pending.values for record_id in record_ids):
                raise ValueError(
                    "Labeling records, but not all " "record_ids were pending."
                )

            data = [
                (int(labels[i]), labeling_times[i], notes[i], int(record_ids[i]))
                for i in range(n_records_labeled)
            ]

            # If not prior, we need to update records.
            query = (
                "UPDATE results SET label=?, labeling_time=?, "
                "notes=? WHERE record_id=?"
            )

        # Add the rows to the database.
        con = self._connect_to_sql()
        cur = con.cursor()
        cur.executemany(query, data)
        con.commit()
        con.close()

    def _add_labeling_data_simulation_mode(self, rows):
        """Add labeling and model data to the results table.

        Add the labeling data and the model data at the same time to the
        results table. This is used for the simulation mode, since the model
        data is available at the time of labeling.

        Arguments
        ----------
        rows : list of tuples
            List of tuples (record_id: int, label: int, classifier: str,
            query_strategy: str, balance_strategy: str, feature_extraction: str,
             training_set: int, labeling_time: int, notes: str).
        """
        query = (
            "INSERT INTO results (record_id, label, classifier, "
            "query_strategy, balance_strategy, feature_extraction, "
            "training_set, labeling_time, notes) "
            "VALUES (?, ?, ?, ?, ?, ?, ?, ?, ?)"
        )

        con = self._connect_to_sql()
        cur = con.cursor()
        cur.executemany(query, rows)
        con.commit()
        con.close()

    def update_decision(self, record_id, label, note=None):
        """Change the label of an already labeled record.

        Arguments
        ---------
        record_id: int
            Id of the record whose label should be changed.
        label: 0 / 1
            New label of the record.
        note: str
            Note to add to the record.
        """

        con = self._connect_to_sql()
        cur = con.cursor()

        # Change the label.
        cur.execute(
            "UPDATE results SET label = ?, notes = ? " "WHERE record_id = ?",
            (label, note, record_id),
        )

        # Add the change to the decision changes table.
        cur.execute(
            (
                "INSERT INTO decision_changes (record_id, new_label, time) "
                "VALUES (?, ?, ?)"
            ),
            (record_id, label, datetime.now()),
        )

        con.commit()
        con.close()

    def delete_record_labeling_data(self, record_id):
        """Delete the labeling data for the given record id.

        Arguments
        ----------
        record_id : str
            Identifier of the record to delete.

        """
        current_time = datetime.now()

        con = self._connect_to_sql()
        cur = con.cursor()
        cur.execute("DELETE FROM results WHERE record_id=?", (record_id,))

        # Add the change to the decision changes table.
        cur.execute(
            (
                "INSERT INTO decision_changes (record_id, new_label, time) "
                "VALUES (?,?, ?)"
            ),
            (record_id, None, current_time),
        )
        con.commit()
        con.close()

    def get_decision_changes(self):
        """Get the record ids for any decision changes.

        Get the record ids of the records whose labels have been changed
        after the original labeling action.

        Returns
        -------
        pd.DataFrame
            Dataframe with columns 'record_id', 'new_label', and 'time' for
            each record of which the labeling decision was changed.
        """
        con = self._connect_to_sql()
        change_table = pd.read_sql_query("SELECT * FROM decision_changes", con)
        con.close()
        return change_table

    def get_record_table(self):
        """Get the record table of the state.

        Returns
        -------
        pd.Series:
            Series with name 'record_id' containing the record ids.
        """
        con = self._connect_to_sql()
        record_table = pd.read_sql_query("SELECT * FROM record_table", con)
        record_table = record_table["record_id"]
        con.close()
        return record_table

    def get_last_probabilities(self):
        """Get the probabilities produced by the last classifier.

        Returns
        -------
        pd.Series:
            Series with name 'proba' containing the probabilities.
        """
        con = self._connect_to_sql()
        last_probabilities = pd.read_sql_query("SELECT * FROM last_probabilities", con)
        con.close()
        return last_probabilities["proba"]

    def get_last_ranking(self):
        """Get the ranking from the state.

        Returns
        -------
        pd.DataFrame
            Dataframe with columns 'record_id', 'ranking', 'classifier',
            'query_strategy', 'balance_strategy', 'feature_extraction',
            'training_set' and 'time'. It has one row for each record in the
            dataset, and is ordered by ranking.
        """
        con = self._connect_to_sql()
        last_ranking = pd.read_sql_query("SELECT * FROM last_ranking", con)
        con.close()
        return last_ranking

    def _move_ranking_data_to_results(self, record_ids):
        """Move data from the ranking to the results table.

        Move the data with the given record_ids from the last_ranking table
        to the results table.

        Arguments
        ---------
        record_ids: list
            List of record ids in last ranking whose model data should be added
            to the results table.
        """
        if self.model_has_trained:
            record_list = [(record_id,) for record_id in record_ids]
            con = self._connect_to_sql()
            cur = con.cursor()
            cur.executemany(
                """INSERT INTO results (record_id, classifier, query_strategy,
                balance_strategy, feature_extraction, training_set)
                SELECT record_id, classifier, query_strategy,
                balance_strategy, feature_extraction, training_set
                FROM last_ranking
                WHERE record_id=?""",
                record_list,
            )
            con.commit()
            con.close()
        else:
            raise StateError("Save trained model data " "before using this function.")

    def query_top_ranked(self, n):
        """Get the top ranked records from the ranking table.

        Get the top n instances from the pool according to the last ranking.
        Add the model data to the results table.

        Arguments
        ---------
        n: int
            Number of instances.

        Returns
        -------
        list
            List of record_ids of the top n ranked records.
        """
        if self.model_has_trained:
            pool = self.get_pool()
            top_n_records = pool[:n].to_list()
            self._move_ranking_data_to_results(top_n_records)
        else:
            raise StateError("Save trained model data " "before using this function.")

        return top_n_records

    # GET FUNCTIONS
    def get_data_by_query_number(self, query, columns=None):
        """Get the data of a specific query from the results table.

        Arguments
        ---------
        query: int
            Number of the query of which you want the data. query=0 corresponds
            to all the prior records.
        columns: list
            List of columns names of the results table.

        Returns
        -------
        pd.DataFrame
            Dataframe containing the data from the results table with the given
            query number and columns.
        """
        if columns is not None:
            if not isinstance(columns, list):
                raise ValueError("The columns argument should be a list.")
        col_query_string = "*" if columns is None else ",".join(columns)

        if query == 0:
            sql_query = (
                f"SELECT {col_query_string} FROM results WHERE "
                f"query_strategy='prior'"
            )
        else:
            rowid = query + self.n_priors
            sql_query = (
                f"SELECT {col_query_string} FROM results WHERE " f"rowid={rowid}"
            )

        con = self._connect_to_sql()
        data = pd.read_sql_query(sql_query, con)
        con.close()
        return data

    def get_data_by_record_id(self, record_id, columns=None):
        """Get the data of a specific query from the results table.

        Arguments
        ---------
        record_id: int
            Record id of which you want the data.
        columns: list
            List of columns names of the results table.

        Returns
        -------
        pd.DataFrame
            Dataframe containing the data from the results table with the given
            record_id and columns.
        """
        query_string = "*" if columns is None else ",".join(columns)

        con = self._connect_to_sql()
        data = pd.read_sql_query(
            f"SELECT {query_string} FROM results WHERE record_id={record_id}", con
        )
        con.close()
        return data

    def get_dataset(self, columns=None, priors=True, pending=False):
        """Get a subset from the results table.

        Can be used to get any column subset from the results table.
        Most other get functions use this one, except some that use a direct
        SQL query for efficiency.

        Arguments
        ---------
        columns: list, str
            List of columns names of the results table, or a string containing
            one column name.
        priors: bool
            Whether to keep the records containing the prior knowledge.
        pending: bool
            Whether to keep the records which are pending a labeling decision.

        Returns
        -------
        pd.DataFrame:
            Dataframe containing the data of the specified columns of the
            results table.
        """
        if isinstance(columns, str):
            columns = [columns]

        if (not priors) or (not pending):
            sql_where = []
            if not priors:
                sql_where.append("query_strategy is not 'prior'")
            if not pending:
                sql_where.append("label is not NULL")

            sql_where_str = f"WHERE {sql_where[0]}"
            if len(sql_where) == 2:
                sql_where_str += f" AND {sql_where[1]}"
        else:
            sql_where_str = ""

        # Query the database.
        query_string = "*" if columns is None else ",".join(columns)
        con = self._connect_to_sql()
        data = pd.read_sql_query(
            f"SELECT {query_string} FROM results {sql_where_str}", con
        )
        con.close()

        return data

    def get_order_of_labeling(self, priors=True, pending=False):
        """Get full array of record id's in order that they were labeled.

        Arguments
        ---------
        priors: bool
            Whether to keep the records containing the prior knowledge.
        pending: bool
            Whether to keep the records are pending a labeling decision.

        Returns
        -------
        pd.Series:
            The record_id's in the order that they were labeled.
        """
        return self.get_dataset("record_id", priors=priors, pending=pending)[
            "record_id"
        ]

    def get_priors(self, columns=["record_id"]):
        """Get the record ids of the priors.

        Returns
        -------
        pd.Series:
            The record_id's of the priors in the order they were added.
        """

        query_string = "*" if columns is None else ",".join(columns)

        con = self._connect_to_sql()
        data = pd.read_sql_query(
            f"SELECT {query_string} FROM results" " WHERE query_strategy is 'prior'",
            con,
        )
        con.close()

        return data

    def get_labels(self, priors=True, pending=False):
        """Get the labels from the state.

        Arguments
        ---------
        priors: bool
            Whether to keep the records containing the prior knowledge.
        pending: bool
            Whether to keep the records which are pending a labeling decision.

        Returns
        -------
        pd.Series:
            Series containing the labels at each labelling moment.
        """

        return self.get_dataset("label", priors=priors, pending=pending)["label"]

    def get_classifiers(self, priors=True, pending=False):
        """Get the classifiers from the state.

        Arguments
        ---------
        priors: bool
            Whether to keep the records containing the prior knowledge.
        pending: bool
            Whether to keep the records which are pending a labeling decision.

        Returns
        -------
        pd.Series:
            Series containing the classifier used at each labeling moment.
        """
        return self.get_dataset("classifier", priors=priors, pending=pending)[
            "classifier"
        ]

    def get_query_strategies(self, priors=True, pending=False):
        """Get the query strategies from the state.

        Arguments
        ---------
        priors: bool
            Whether to keep the records containing the prior knowledge.
        pending: bool
            Whether to keep the records which are pending a labeling decision.

        Returns
        -------
        pd.Series:
            Series containing the query strategy used to get the record to
            query at each labeling moment.
        """
        return self.get_dataset("query_strategy", priors=priors, pending=pending)[
            "query_strategy"
        ]

    def get_balance_strategies(self, priors=True, pending=False):
        """Get the balance strategies from the state.

        Arguments
        ---------
        priors: bool
            Whether to keep the records containing the prior knowledge.
        pending: bool
            Whether to keep the records which are pending a labeling decision.

        Returns
        -------
        pd.Series:
            Series containing the balance strategy used to get the training
            data at each labeling moment.
        """
        return self.get_dataset("balance_strategy", priors=priors, pending=pending)[
            "balance_strategy"
        ]

    def get_feature_extraction(self, priors=True, pending=False):
        """Get the query strategies from the state.

        Arguments
        ---------
        priors: bool
            Whether to keep the records containing the prior knowledge.
        pending: bool
            Whether to keep the records which are pending a labeling decision.

        Returns
        -------
        pd.Series:
            Series containing the feature extraction method used for the
            classifier input at each labeling moment.
        """
        return self.get_dataset("feature_extraction", priors=priors, pending=pending)[
            "feature_extraction"
        ]

    def get_training_sets(self, priors=True, pending=False):
        """Get the training_sets from the state.

        Arguments
        ---------
        priors: bool
            Whether to keep the records containing the prior knowledge.
        pending: bool
            Whether to keep the records which are pending a labeling decision.

        Returns
        -------
        pd.Series:
            Series containing the training set on which the classifier was fit
            at each labeling moment.
        """
        return self.get_dataset("training_set", priors=priors, pending=pending)[
            "training_set"
        ]

    def get_labeling_times(self, time_format="int", priors=True, pending=False):
        """Get the time of labeling from the state.

        Arguments
        ---------
        time_format: 'int' or 'datetime'
            Format of the return value. If it is 'int' you get a UTC timestamp,
            if it is 'datetime' you get datetime instead of an integer.
        priors: bool
            Whether to keep the records containing the prior knowledge.
        pending: bool
            Whether to keep the records which are pending a labeling decision.

        Returns
        -------
        pd.Series:
            If format='int' you get a UTC timestamp (integer number of
            microseconds), if it is 'datetime' you get datetime format.
        """
        times = self.get_dataset("labeling_time", priors=priors, pending=pending)[
            "labeling_time"
        ]

        # Convert time to datetime format.
        if time_format == "datetime":
            times = times.applymap(lambda x: datetime.utcfromtimestamp(x / 10**6))

        return times

    # Get pool, labeled and pending in slightly more optimized way than via
    # get_dataset.
    def get_pool(self):
        """Get the unlabeled, not-pending records in ranking order.

        Get the pool of unlabeled records, not pending a labeling decision,
        in the ranking order. If you only want the records in the pool, this
        is more efficient than via 'get_pool_labeled_pending'.

        Returns
        -------
        pd.Series
            Series containing the record_ids of the unlabeled, not pending
            records, in the order of the last available ranking.
        """
        # If model has trained, using ranking to order pool.
        con = self._connect_to_sql()
        if self.model_has_trained:
            query = """SELECT last_ranking.record_id, last_ranking.ranking,
                    results.query_strategy
                    FROM last_ranking
                    LEFT JOIN results
                    ON last_ranking.record_id = results.record_id
                    WHERE results.query_strategy is null
                    ORDER BY ranking
                    """
            df = pd.read_sql_query(query, con)

        # Else return all records not yet in the results table.
        else:
            query = """SELECT record_table.record_id, results.query_strategy
                    FROM record_table
                    LEFT JOIN results
                    ON record_table.record_id = results.record_id
                    WHERE results.query_strategy is null
                    """
            df = pd.read_sql_query(query, con)

        con.close()
        return df["record_id"]

    def get_labeled(self):
        """Get the labeled records in order of labeling.

        Get the record_ids and labels of the labeled records in order of
        labeling. If you only want the labeled records, this is more efficient
        than via 'get_pool_labeled_pending'.

        Returns
        -------
        pd.DataFrame
            Dataframe containing the record_ids and labels of the labeled
            records, in the order that they were labeled.
        """
        con = self._connect_to_sql()
        query = """SELECT record_id, label FROM results
         WHERE label is not null"""
        df = pd.read_sql_query(query, con)
        con.close()
        return df

    def get_pending(self):
        """Get the record_ids of the records pending a labeling decision.

        If you only want the pending records, this is more efficient
        than via 'get_pool_labeled_pending'.

        Returns
        -------
        pd.Series
            A series containing the record_ids of the records whose label is
            pending.
        """
        con = self._connect_to_sql()
        query = """SELECT record_id FROM results WHERE label is null"""
        df = pd.read_sql_query(query, con)
        con.close()
        return df["record_id"]

    def get_pool_labeled_pending(self):
        """Return the unlabeled pool, labeled and pending records.

        Convenience function to get the pool, labeled and pending records in
        one SQL query. If you only want one of these, it is more efficient to
        use the methods 'get_pool', 'get_labeled' or 'get_pending'.

        Returns
        -------
        tuple (pd.Series, pd.DataFrame, pd.Series):
            Returns a tuple (pool, labeled, pending). Pool is a series
            containing the unlabeled, not pending record_ids, ordered by the
            last predicted ranking of the model. Labeled is a dataframe
            containing the record_ids and labels of the labeled records, in the
            order that they were labeled. Pending is a series containing the
            record_ids of the records whose label is pending.
        """
        con = self._connect_to_sql()

        query = """SELECT record_table.record_id, results.label,
                results.rowid AS label_order, results.query_strategy,
                last_ranking.ranking
                FROM record_table
                LEFT JOIN results
                ON results.record_id=record_table.record_id
                LEFT JOIN last_ranking
                ON record_table.record_id=last_ranking.record_id
                ORDER BY label_order, ranking
                """

        df = pd.read_sql_query(query, con)
        con.close()
        labeled = df.loc[~df["label"].isna()].loc[:, ["record_id", "label"]].astype(int)
        pool = df.loc[df["label_order"].isna(), "record_id"].astype(int)
        pending = (
            df.loc[df["label"].isna() & ~df["query_strategy"].isna()]
            .loc[:, "record_id"]
            .astype(int)
        )

        return pool, labeled, pending



# Copyright 2019-2022 The ASReview Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging

import numpy as np
import pandas as pd

from asreview.config import COLUMN_DEFINITIONS
from asreview.config import LABEL_NA
from asreview.exceptions import BadFileFormatError
from asreview.utils import _entry_points


def type_from_column(col_name, col_definitions):
    """Transform a column name to its standardized form.

    Arguments
    ---------
    col_name: str
        Name of the column in the dataframe.
    col_definitions: dict
        Dictionary of {standardized_name: [list of possible names]}.
        Ex. {"title": ["title", "primary_title"],
            "authors": ["authors", "author names", "first_authors"]}

    Returns
    -------
    str:
        The standardized name. If it wasn't found, return None.
    """
    for name, definition in col_definitions.items():
        if col_name.lower() in definition:
            return name
    return None


def convert_keywords(keywords):
    """Split keywords separated by commas etc to lists."""
    if not isinstance(keywords, str):
        return keywords

    current_best = [keywords]
    for splitter in [", ", "; ", ": ", ";", ":"]:
        new_split = keywords.split(splitter)
        if len(new_split) > len(current_best):
            current_best = new_split
    return current_best


def _is_record_id_unique(s):
    if len(pd.unique(s)) != len(s.index):
        raise ValueError("Column 'record_id' contains duplicate values.")


def _is_record_id_notnull(s):
    if s.isnull().any():
        raise ValueError("Column 'record_id' contains missing values.")


def _is_record_id_int(s):
    try:
        pd.to_numeric(s).astype(int)
    except Exception:
        raise ValueError("Column 'record_id' should contain integer values.")


def _standardize_dataframe(df, column_def={}):
    """Create a ASReview readable dataframe.

    The main purpose is to rename columns with slightly different names;
    'authors' vs 'first_authors', etc. This greatly widens the compatibility
    with different datasets.

    Arguments
    ---------
    df: pandas.DataFrame
        Unclean dataframe to be cleaned up.

    Returns
    -------
    pd.DataFrame:
        Cleaned dataframe with proper column names.
    """
    all_column_spec = {}

    # remove whitespace from colnames
    df.columns = df.columns.str.strip()

    # map columns on column specification
    col_names = list(df)
    for column_name in col_names:
        # First try the custom column definitions if supplied.
        data_type = type_from_column(column_name, column_def)
        if data_type is not None:
            all_column_spec[data_type] = column_name
            continue
        # Then try the standard specifications in ASReview.
        data_type = type_from_column(column_name, COLUMN_DEFINITIONS)
        if data_type is not None:
            all_column_spec[data_type] = column_name

    # Check if we either have abstracts or titles.
    col_names = list(all_column_spec)
    if "abstract" not in col_names and "title" not in col_names:
        raise BadFileFormatError(
            "File supplied without 'abstract' or 'title'" " fields."
        )
    if "abstract" not in col_names:
        logging.warning("Unable to detect abstracts in dataset.")
    if "title" not in col_names:
        logging.warning("Unable to detect titles in dataset.")

    # Replace NA values with empty strings.
    for col in ["title", "abstract", "authors", "keywords", "notes"]:
        try:
            df[all_column_spec[col]] = np.where(
                pd.isnull(df[all_column_spec[col]]),
                "",
                df[all_column_spec[col]].astype(str),
            )
        except KeyError:
            pass

    # Convert labels to integers.
    if "included" in col_names:
        try:
            col = all_column_spec["included"]
            df[col].fillna(LABEL_NA, inplace=True)
            df[col] = pd.to_numeric(df[col])
        except KeyError:
            pass
        except ValueError:
            logging.warning(
                "Failed to parse label column name, no labels will" " be present."
            )
            df.rename(columns={"label": "included"})
            all_column_spec.pop("included")

    # TODO: Make sure 'record_id' column in original dataset does not get overwritten.
    # # If the we have a record_id (for example from an ASReview export) use it.
    # if "record_id" in list(df):
    #
    #     # validate record_id column
    #     _is_record_id_notnull(df["record_id"])
    #     _is_record_id_unique(df["record_id"])
    #     _is_record_id_int(df["record_id"])
    #
    # # Create a new index if we haven't found it in the data.
    # else:
    #     df["record_id"] = np.arange(len(df.index))
    df["record_id"] = np.arange(len(df.index)).astype('int64')

    # set the index
    df.set_index("record_id", inplace=True)

    return df, all_column_spec


def list_readers():
    """List available dataset reader classes.

    Returns
    -------
    list:
        Classes of available dataset readers in alphabetical order.
    """
    return [e.load() for e in _entry_points(group="asreview.readers")]


def list_writers():
    """List available dataset writer classes.

    Returns
    -------
    list:
        Classes of available dataset writers in alphabetical order.
    """
    return [e.load() for e in _entry_points(group="asreview.writers")]


def get_reader_class(name):
    """Get class of dataset reader from string.

    Arguments
    ---------
    name: str
        Name of the dataset reader, e.g. '.csv', '.tsv' or '.xlsx'.

    Returns
    -------
    class:
        Class corresponding to the name.
    """
    return _entry_points(group="asreview.readers")[name].load()


def get_writer_class(name):
    """Get class of dataset writer from string.

    Arguments
    ---------
    name: str
        Name of the dataset writer, e.g. '.csv', '.tsv' or '.xlsx'.

    Returns
    -------
    class:
        Class corresponding to the name.
    """

    return _entry_points(group="asreview.writers")[name].load()



# Copyright 2019-2022 The ASReview Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Max sampling while saving prediction probabilities."""

import numpy as np

from asreview.models.query.base import ProbaQueryStrategy


class MaxQuery(ProbaQueryStrategy):
    """Maximum query strategy (``max``).

    Choose the most likely samples to be included according to the model.
    """

    name = "max"
    label = "Maximum"

    def _query(self, predictions, n_instances, X=None):
        query_indices = np.argsort(predictions[:, 0])[:n_instances]

        return query_indices



import random
import time
from io import BytesIO
from typing import Union
from urllib.request import urlopen

from flask.testing import FlaskClient

import asreview.webapp.tests.utils.crud as crud
import asreview.webapp.tests.utils.misc as misc
from asreview.project import ASReviewProject
from asreview.webapp.authentication.models import Project
from asreview.webapp.tests.utils.config_parser import all_users
from asreview.webapp.tests.utils.config_parser import get_user
from asreview.webapp.tests.utils.misc import get_project_id


def process_response(response):
    """Breaks response in a tuple containing the status code
    and json data."""
    return (response.status_code, response.json)


# ########################
# General API calls
# ########################


def call_root_url(client):
    response = client.get("/")
    status_code, data = process_response(response)
    return (status_code, data, response.text)


def call_boot_url(client):
    response = client.get("/boot")
    return process_response(response)


# ########################
# Authentication API calls
# ########################


def signin_user(client, user):
    """Signs in a user through the api"""
    # If a password is not set, we need to get it
    if not hasattr(user, "password"):
        users = all_users()
        user.password = users[user.identifier].password
    # request
    response = client.post(
        "/auth/signin", data={"email": user.identifier, "password": user.password}
    )
    return process_response(response)


def signup_user(client, user):
    """Signs up a user through the api"""
    response = client.post(
        "/auth/signup",
        data={
            "identifier": user.email,
            "email": user.email,
            "name": user.name,
            "password": user.password,
            "affiliation": user.affiliation,
            "origin": "asreview",
        },
    )
    return process_response(response)


def signout_user(client):
    """Sign out user"""
    response = client.delete("/auth/signout")
    return process_response(response)


def confirm_user(client, user):
    response = client.post(
        "/auth/confirm_account", data={"user_id": user.id, "token": user.token}
    )
    return process_response(response)


def forgot_password(client, user):
    response = client.post("/auth/forgot_password", data={"email": user.email})
    return process_response(response)


def reset_password(client, user):
    response = client.post(
        "/auth/reset_password",
        data={"password": user.password, "token": user.token, "user_id": user.id},
    )
    return process_response(response)


def update_user(client, data):
    response = client.post("/auth/update_profile", data=data)
    return process_response(response)


def refresh(client):
    response = client.get("/auth/refresh")
    return process_response(response)


def get_profile(client: FlaskClient):
    response = client.get("/auth/get_profile")
    return process_response(response)


# ########################
# Teams API calls
# ########################


def invite(client, project, user):
    url = f"/api/invitations/projects/{get_project_id(project)}/users/{user.id}"
    response = client.post(url)
    return process_response(response)


def list_invitations(client):
    response = client.get("/api/invitations")
    return process_response(response)


def list_collaborators(client, project):
    response = client.get(f"/api/projects/{get_project_id(project)}/users")
    return process_response(response)


def accept_invitation(client, project):
    response = client.post(
        f"/api/invitations/projects/{get_project_id(project)}/accept", data={}
    )
    return process_response(response)


def reject_invitation(client, project):
    response = client.delete(
        f"/api/invitations/projects/{get_project_id(project)}/reject", data={}
    )
    return process_response(response)


def delete_invitation(client, project, user):
    response = client.delete(
        f"/api/invitations/projects/{get_project_id(project)}/users/{user.id}", data={}
    )
    return process_response(response)


def delete_collaboration(client, project, user):
    response = client.delete(
        f"/api/projects/{get_project_id(project)}/users/{user.id}", data={}
    )
    return process_response(response)


# ########################
# Project API calls
# ########################


def get_all_projects(client: FlaskClient):
    response = client.get("/api/projects")
    return process_response(response)


def create_project(
    client: FlaskClient,
    project_name: str,
    mode: str = "explore",
    authors: str = "authors",
    description: str = "description",
):
    response = client.post(
        "/api/projects/info",
        data={
            "mode": mode,
            "name": project_name,
            "authors": authors,
            "description": description,
        },
    )
    return process_response(response)


def create_project_from_dict(client: FlaskClient, data: dict):
    response = client.post(
        "/api/projects/info",
        data=data,
    )
    return process_response(response)


def update_project(
    client: FlaskClient,
    project: Union[Project, ASReviewProject],
    name: str = "name",
    mode: str = "explore",
    authors: str = "authors",
    description: str = "description",
):
    response = client.put(
        f"/api/projects/{get_project_id(project)}/info",
        data={
            "mode": mode,
            "name": name,
            "authors": authors,
            "description": description,
        },
    )
    return process_response(response)


def upgrade_project(client: FlaskClient, project: Union[Project, ASReviewProject]):
    response = client.get(f"/api/projects/{get_project_id(project)}/upgrade_if_old")
    return process_response(response)


def import_project(client: FlaskClient, url: str):
    with urlopen(url) as project_file:
        response = client.post(
            "/api/projects/import_project",
            data={"file": (BytesIO(project_file.read()), "project.asreview")},
        )
    return process_response(response)


def get_project_stats(client: FlaskClient):
    response = client.get("/api/projects/stats")
    return process_response(response)


def get_demo_data(client: FlaskClient, subset: str):
    response = client.get(f"/api/datasets?subset={subset}")
    return process_response(response)


def upload_data_to_project(
    client: FlaskClient, project: Union[Project, ASReviewProject], data: dict
):
    response = client.post(
        f"/api/projects/{get_project_id(project)}/data",
        data=data,
    )
    return process_response(response)


def get_project_data(client: FlaskClient, project: Union[Project, ASReviewProject]):
    response = client.get(f"/api/projects/{get_project_id(project)}/data")
    return process_response(response)


def get_project_dataset_writer(
    client: FlaskClient, project: Union[Project, ASReviewProject]
):
    response = client.get(f"/api/projects/{get_project_id(project)}/dataset_writer")
    return process_response(response)


def search_project_data(
    client: FlaskClient, project: Union[Project, ASReviewProject], query: str
):
    response = client.get(f"/api/projects/{get_project_id(project)}/search?q={query}")
    return process_response(response)


def get_prior_random_project_data(
    client: FlaskClient, project: Union[Project, ASReviewProject]
):
    response = client.get(f"/api/projects/{get_project_id(project)}/prior_random")
    return process_response(response)


def label_random_project_data_record(
    client: FlaskClient, project: Union[Project, ASReviewProject], label: int
):
    # get random data
    _, data = get_prior_random_project_data(client, project)
    # select a specific record
    record = random.choice(data["result"])
    doc_id = record["id"]
    return label_project_record(client, project, doc_id, label, note="")


def label_project_record(
    client: FlaskClient,
    project: Union[Project, ASReviewProject],
    doc_id: int,
    label: str,
    prior: int = 1,
    note: str = "",
):
    response = client.post(
        f"/api/projects/{get_project_id(project)}/record/{doc_id}",
        data={"doc_id": doc_id, "label": label, "is_prior": prior, "note": note},
    )
    return process_response(response)


def update_label_project_record(
    client: FlaskClient,
    project: Union[Project, ASReviewProject],
    doc_id: int,
    label: str,
    prior: int = 1,
    note: str = "",
):
    response = client.put(
        f"/api/projects/{get_project_id(project)}/record/{doc_id}",
        data={"doc_id": doc_id, "label": label, "is_prior": prior, "note": note},
    )
    return process_response(response)


def get_labeled_project_data(
    client: FlaskClient, project: Union[Project, ASReviewProject]
):
    response = client.get(f"/api/projects/{get_project_id(project)}/labeled")
    return process_response(response)


def get_labeled_project_data_stats(
    client: FlaskClient, project: Union[Project, ASReviewProject]
):
    response = client.get(f"/api/projects/{get_project_id(project)}/labeled_stats")
    return process_response(response)


def get_project_algorithms_options(client: FlaskClient):
    response = client.get("/api/algorithms")
    return process_response(response)


def set_project_algorithms(
    client: FlaskClient, project: Union[Project, ASReviewProject], data: dict
):
    response = client.post(
        f"/api/projects/{get_project_id(project)}/algorithms", data=data
    )
    return process_response(response)


def get_project_algorithms(
    client: FlaskClient, project: Union[Project, ASReviewProject]
):
    response = client.get(f"/api/projects/{get_project_id(project)}/algorithms")
    return process_response(response)


def start_project_algorithms(
    client: FlaskClient, project: Union[Project, ASReviewProject]
):
    response = client.post(f"/api/projects/{get_project_id(project)}/start")
    return process_response(response)


def get_project_status(client: FlaskClient, project: Union[Project, ASReviewProject]):
    response = client.get(f"/api/projects/{get_project_id(project)}/status")
    return process_response(response)


def set_project_status(
    client: FlaskClient, project: Union[Project, ASReviewProject], status: str
):
    response = client.put(
        f"/api/projects/{get_project_id(project)}/status", data={"status": status}
    )
    return process_response(response)


def export_project_dataset(
    client: FlaskClient, project: Union[Project, ASReviewProject], format: str
):
    id = get_project_id(project)
    response = client.get(f"/api/projects/{id}/export_dataset?file_format={format}")
    return process_response(response)


def export_project(
    client: FlaskClient,
    project: Union[Project, ASReviewProject],
):
    response = client.get(f"/api/projects/{get_project_id(project)}/export_project")
    return process_response(response)


def get_project_progress(
    client: FlaskClient,
    project: Union[Project, ASReviewProject],
):
    response = client.get(f"/api/projects/{get_project_id(project)}/progress")
    return process_response(response)


def get_project_progress_density(
    client: FlaskClient,
    project: Union[Project, ASReviewProject],
):
    response = client.get(f"/api/projects/{get_project_id(project)}/progress_density")
    return process_response(response)


def get_project_progress_recall(
    client: FlaskClient,
    project: Union[Project, ASReviewProject],
):
    response = client.get(f"/api/projects/{get_project_id(project)}/progress_recall")
    return process_response(response)


def get_project_current_document(
    client: FlaskClient,
    project: Union[Project, ASReviewProject],
):
    response = client.get(f"/api/projects/{get_project_id(project)}/get_document")
    return process_response(response)


def delete_project(
    client: FlaskClient,
    project: Union[Project, ASReviewProject],
):
    response = client.delete(f"/api/projects/{get_project_id(project)}/delete")
    return process_response(response)


# ########################
# General procedures
# ########################


def create_and_signin_user(client, test_user_id=1):
    """Creates a user account and signs in with that account."""
    # signup user
    user = get_user(test_user_id)
    signup_user(client, user)
    # refresh user
    stored_user = crud.get_user_by_identifier(user.identifier)
    # signin user
    signin_user(client, user)
    # return the user
    return stored_user


def upload_label_set_and_start_model(client, project, dataset):
    """Uploads a dataset to a created project and adds and starts
    a random model."""
    # upload dataset
    upload_data_to_project(client, project, data=dataset)
    # label 2 random records
    label_random_project_data_record(client, project, 1)
    label_random_project_data_record(client, project, 0)
    # select a model
    model_data = misc.choose_project_algorithms()
    set_project_algorithms(client, project, data=model_data)
    # start the model
    start_project_algorithms(client, project)
    # make sure model is done
    time.sleep(10)



# Copyright 2019-2022 The ASReview Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import argparse
from pathlib import Path

import pandas as pd

from asreview.entry_points.base import BaseEntryPoint
from asreview.project import get_project_path
from asreview.project import open_state


def _parse_state_inspect_args():
    # parse arguments if available
    parser = argparse.ArgumentParser(
        prog="state-inspect", description="Inspect state file."
    )
    parser.add_argument(
        "project_id", type=str, help="Project_id or path to ASReview file."
    )
    parser.add_argument(
        "table",
        type=str,
        help="Table to view (e.g. results, record_table, last_ranking).",
    )

    return parser


class StateInspectEntryPoint(BaseEntryPoint):
    """Entry point to inspect ASReview LAB review progress."""

    def execute(self, argv):
        parser = _parse_state_inspect_args()
        args = parser.parse_args(argv)

        if Path(args.project_id).suffix == ".asreview":
            project_path = args.project_id
        else:
            project_path = get_project_path(args.project_id)

        with open_state(project_path) as s:
            conn = s._connect_to_sql()

        df = pd.read_sql(f"select * from {args.table}", conn)

        if args.table == "results":
            df["label"] = df["label"].astype(pd.Int64Dtype())

        print(f"Table '{args.table}':\n")
        print(df)
        print("\n")



import inspect
import time
from typing import Union

import pytest
from flask.testing import FlaskClient

import asreview.webapp.tests.utils.api_utils as au
import asreview.webapp.tests.utils.crud as crud
import asreview.webapp.tests.utils.misc as misc
from asreview.project import ASReviewProject
from asreview.webapp import DB
from asreview.webapp.authentication.models import Project
from asreview.webapp.tests.utils.misc import current_app_is_authenticated
from asreview.webapp.tests.utils.misc import retrieve_project_url_github

# NOTE: I don't see a plugin that can be used for testing
# purposes
UPLOAD_DATA = [
    {"benchmark": "synergy:van_der_Valk_2021"},
    {
        "url": "https://raw.githubusercontent.com/asreview/"
        + "asreview/master/tests/demo_data/generic_labels.csv"
    },
]
IMPORT_PROJECT_URLS = retrieve_project_url_github()

# NOTE: the setup fixture entails: a FlaskClient, 1 user (signed in),
# and a project of this user OR a project from an unauthenticated app.
# The fixture is parametrized! It runs the authenticated app and the
# unauthenticated app.


# Test getting all projects
def test_get_projects(setup):
    client, user1, project = setup
    status_code, data = au.get_all_projects(client)
    assert status_code == 200
    assert len(data["result"]) == 1
    found_project = data["result"][0]
    if current_app_is_authenticated():
        assert found_project["id"] == project.project_id
        assert found_project["owner_id"] == user1.id
    else:
        assert found_project["id"] == project.config["id"]


# Test create a project
def test_create_projects(setup):
    client, _, _ = setup
    project_name = "new_project"

    status_code, data = au.create_project(client, project_name)
    assert status_code == 201
    assert data["name"] == project_name


# Test upgrading a post v0.x project
def test_try_upgrade_a_modern_project(setup):
    client, _, project = setup
    # verify version
    data = misc.read_project_file(project)
    assert not data["version"].startswith("0")

    status_code, data = au.upgrade_project(client, project)
    assert status_code == 400
    assert data["message"] == "Can only convert v0.x projects."


# Test upgrading a v0.x project
def test_upgrade_an_old_project(setup):
    client, user, _ = setup
    # get an old version from github
    old_project_url = retrieve_project_url_github("v0.19")
    project = misc.copy_github_project_into_asreview_folder(old_project_url)
    # we need to make sure this new, old-style project can be found
    # under current user if the app is authenticated
    if current_app_is_authenticated():
        new_project = Project(project_id=project.config.get("id"))
        project = crud.create_project(DB, user, new_project)
    print(type(project))
    # try to convert
    status_code, data = au.upgrade_project(client, project)
    assert status_code == 200
    assert data["success"]


# Test importing old projects, verify ids
@pytest.mark.parametrize("url", IMPORT_PROJECT_URLS)
def test_import_project_files(setup, url):
    client, user, first_project = setup
    # import project
    status_code, data = au.import_project(client, url)
    # get contents asreview folder
    folders = set(misc.get_folders_in_asreview_path())
    # asserts
    assert len(folders) == 2
    assert status_code == 200
    assert isinstance(data, dict)
    if current_app_is_authenticated():
        # assert it exists in the database
        assert crud.count_projects() == 2
        project = crud.last_project()
        assert data["id"] != first_project.project_id
        assert data["id"] == project.project_id
        # assert the owner is current user
        assert data["owner_id"] == user.id
    else:
        assert data["id"] != first_project.config.get("id")
    # in auth/non-auth the project folder must exist in the asreview folder
    assert data["id"] in set([f.stem for f in folders])


# Test get stats in setup state
def test_get_projects_stats_setup_stage(setup):
    client, _, _ = setup
    status_code, data = au.get_project_stats(client)
    assert status_code == 200
    assert isinstance(data["result"], dict)
    assert data["result"]["n_in_review"] == 0
    assert data["result"]["n_finished"] == 0
    assert data["result"]["n_setup"] == 1


# Test get stats in review state
def test_get_projects_stats_review_stage(setup):
    client, _, project = setup
    # start the show
    au.upload_label_set_and_start_model(client, project, UPLOAD_DATA[0])
    # get stats
    status_code, data = au.get_project_stats(client)
    assert status_code == 200
    assert isinstance(data["result"], dict)
    assert data["result"]["n_in_review"] == 1
    assert data["result"]["n_finished"] == 0
    assert data["result"]["n_setup"] == 0


# Test get stats in finished state
def test_get_projects_stats_finished_stage(setup):
    client, _, project = setup
    # start the show
    au.upload_label_set_and_start_model(client, project, UPLOAD_DATA[0])
    # manually finish the project
    au.set_project_status(client, project, "finished")
    # get stats
    status_code, data = au.get_project_stats(client)
    assert status_code == 200
    assert isinstance(data["result"], dict)
    assert data["result"]["n_in_review"] == 0
    assert data["result"]["n_finished"] == 1
    assert data["result"]["n_setup"] == 0


# Test known demo data
@pytest.mark.parametrize("subset", ["plugin", "benchmark"])
def test_demo_data_project(setup, subset):
    client, _, _ = setup
    status_code, data = au.get_demo_data(client, subset)
    assert status_code == 200
    assert isinstance(data["result"], list)


# Test unknown demo data
def test_unknown_demo_data_project(setup):
    client, _, _ = setup
    status_code, data = au.get_demo_data(client, "abcdefg")
    assert status_code == 400
    assert data["message"] == "demo-data-loading-failed"


# Test uploading benchmark data to a project
@pytest.mark.parametrize("upload_data", UPLOAD_DATA)
def test_upload_benchmark_data_to_project(setup, upload_data):
    client, _, project = setup
    status_code, data = au.upload_data_to_project(client, project, data=upload_data)
    assert status_code == 200
    if current_app_is_authenticated():
        assert data["project_id"] == project.project_id
    else:
        assert data["project_id"] == project.config.get("id")


# Test getting the data after an upload
@pytest.mark.parametrize("upload_data", UPLOAD_DATA)
def test_get_project_data(setup, upload_data):
    client, _, project = setup
    au.upload_data_to_project(client, project, data=upload_data)
    status_code, data = au.get_project_data(client, project)
    assert status_code == 200
    assert data["filename"] == misc.extract_filename_stem(upload_data)


# Test get dataset writer
def test_get_dataset_writer(setup):
    client, _, project = setup
    # upload data
    au.upload_data_to_project(client, project, data=UPLOAD_DATA[0])
    # get dataset writer
    status_code, data = au.get_project_dataset_writer(client, project)
    assert status_code == 200
    assert isinstance(data["result"], list)


# Test updating a project
def test_update_project_info(setup):
    client, _, project = setup
    # update data
    new_mode = "oracle"
    new_name = "new name"
    new_authors = "new authors"
    new_description = "new description"
    # request
    status_code, data = au.update_project(
        client,
        project,
        name=new_name,
        mode=new_mode,
        authors=new_authors,
        description=new_description,
    )
    assert status_code == 200
    assert data["authors"] == new_authors
    assert data["description"] == new_description
    assert data["mode"] == new_mode
    assert data["name"] == new_name


# Test search data
def test_search_data(setup):
    client, _, project = setup
    # upload dataset
    au.upload_data_to_project(client, project, data=UPLOAD_DATA[0])
    # search
    status_code, data = au.search_project_data(
        client, project, query="Software&n_max=10"
    )
    assert status_code == 200
    assert "result" in data
    assert isinstance(data["result"], list)
    assert len(data["result"]) <= 10


# Test get a selection of random papers to find exclusions
def test_random_prior_papers(setup):
    client, _, project = setup
    # upload dataset
    au.upload_data_to_project(client, project, data=UPLOAD_DATA[0])
    # get random selection
    status_code, data = au.get_prior_random_project_data(client, project)
    assert status_code == 200
    assert "result" in data
    assert isinstance(data["result"], list)
    assert len(data["result"]) > 0


# Test labeling of prior data
@pytest.mark.parametrize("label", [0, 1])
def test_label_item(setup, label):
    client, _, project = setup
    # upload dataset
    au.upload_data_to_project(client, project, data=UPLOAD_DATA[0])
    # label
    status_code, data = au.label_random_project_data_record(client, project, label)
    assert status_code == 200
    assert data["success"]


# Test getting labeled records
def test_get_labeled_project_data(setup):
    client, _, project = setup
    # upload dataset
    au.upload_data_to_project(client, project, data=UPLOAD_DATA[0])
    # label a random record
    au.label_random_project_data_record(client, project, 1)
    # collect labeled records
    status_code, data = au.get_labeled_project_data(client, project)
    assert status_code == 200
    assert "result" in data
    assert isinstance(data["result"], list)
    assert len(data["result"]) == 1


# Test getting labeled records stats
def test_get_labeled_stats(setup):
    client, _, project = setup
    # upload dataset
    au.upload_data_to_project(client, project, data=UPLOAD_DATA[0])
    # label 2 random records
    au.label_random_project_data_record(client, project, 1)
    au.label_random_project_data_record(client, project, 0)
    # collect stats
    status_code, data = au.get_labeled_project_data_stats(client, project)

    assert status_code == 200
    assert isinstance(data, dict)
    assert data["n"] == 2
    assert data["n_exclusions"] == 1
    assert data["n_inclusions"] == 1
    assert data["n_prior"] == 2


# Test listing the available algorithms
def test_list_algorithms(setup):
    client, _, _ = setup
    status_code, data = au.get_project_algorithms_options(client)
    assert status_code == 200
    expected_keys = [
        "balance_strategy",
        "classifier",
        "feature_extraction",
        "query_strategy",
    ]
    for key in expected_keys:
        assert key in data.keys()
        assert isinstance(data[key], list)
        for item in data[key]:
            assert "name" in item.keys()
            assert "label" in item.keys()


# Test setting the algorithms
def test_set_project_algorithms(setup):
    client, _, project = setup
    data = misc.choose_project_algorithms()
    status_code, data = au.set_project_algorithms(client, project, data=data)
    assert status_code == 200
    assert data["success"]


# Test getting the project algorithms
def test_get_project_algorithms(setup):
    client, _, project = setup
    data = misc.choose_project_algorithms()
    au.set_project_algorithms(client, project, data=data)
    # get the project algorithms
    status_code, resp_data = au.get_project_algorithms(client, project)
    assert status_code == 200
    assert resp_data["balance_strategy"] == data["balance_strategy"]
    assert resp_data["feature_extraction"] == data["feature_extraction"]
    assert resp_data["model"] == data["model"]
    assert resp_data["query_strategy"] == data["query_strategy"]


# Test starting the model
def test_start_and_model_ready(setup):
    client, _, project = setup
    # upload dataset
    au.upload_data_to_project(client, project, data=UPLOAD_DATA[0])
    # label 2 random records
    au.label_random_project_data_record(client, project, 1)
    au.label_random_project_data_record(client, project, 0)
    # select a model
    data = misc.choose_project_algorithms()
    au.set_project_algorithms(client, project, data=data)
    # start the model
    status_code, data = au.start_project_algorithms(client, project)
    assert status_code == 200
    assert data["success"]
    # make sure model is done
    time.sleep(10)


# Test status of project
@pytest.mark.parametrize(
    ("state_name", "expected_state"),
    [
        ("creation", None),
        ("setup", "setup"),
        ("review", "review"),
        ("finish", "finished"),
    ],
)
def test_status_project(setup, state_name, expected_state):
    client, _, project = setup
    # call these progression steps
    if state_name in ["setup", "review", "finish"]:
        # upload dataset
        au.upload_data_to_project(client, project, data=UPLOAD_DATA[0])
        # label 2 records
        au.label_random_project_data_record(client, project, 1)
        au.label_random_project_data_record(client, project, 0)
        # select a model
        data = misc.choose_project_algorithms()
        au.set_project_algorithms(client, project, data=data)
    if state_name in ["review", "finish"]:
        # start the model
        au.start_project_algorithms(client, project)
        time.sleep(15)
    if state_name == "finish":
        # mark project as finished
        au.set_project_status(client, project, "finished")

    status_code, data = au.get_project_status(client, project)
    assert status_code == 200
    assert data["status"] == expected_state


# Test exporting the results
@pytest.mark.parametrize("format", ["csv", "tsv", "xlsx"])
def test_export_result(setup, format):
    client, _, project = setup
    # upload dataset
    au.upload_data_to_project(client, project, data=UPLOAD_DATA[0])
    au.label_random_project_data_record(client, project, 1)
    au.label_random_project_data_record(client, project, 0)
    # request
    status_code, _ = au.export_project_dataset(client, project, format)
    assert status_code == 200


# Test exporting the entire project
def test_export_project(setup):
    client, _, project = setup
    # upload dataset
    au.upload_data_to_project(client, project, data=UPLOAD_DATA[0])
    au.label_random_project_data_record(client, project, 1)
    au.label_random_project_data_record(client, project, 0)
    # request
    status_code, _ = au.export_project(client, project)
    assert status_code == 200


# Test setting the project status
@pytest.mark.parametrize("status", ["review", "finished"])
def test_set_project_status(setup, status):
    client, _, project = setup
    # start the show
    au.upload_label_set_and_start_model(client, project, UPLOAD_DATA[0])
    # when setting the status to "review", the project must have another
    # status then "review"
    if status == "review":
        au.set_project_status(client, project, "finished")
    # set project status
    status_code, data = au.set_project_status(client, project, status)
    assert status_code == 200
    assert data["success"]


# Test get progress info
def test_get_progress_info(setup):
    client, _, project = setup
    # upload dataset
    au.upload_data_to_project(client, project, data=UPLOAD_DATA[0])
    # label 2 random records
    au.label_random_project_data_record(client, project, 1)
    au.label_random_project_data_record(client, project, 0)
    # get progress
    status_code, data = au.get_project_progress(client, project)
    assert status_code == 200
    assert isinstance(data, dict)
    assert data["n_excluded"] == 1
    assert data["n_included"] == 1
    assert data["n_pool"] == data["n_papers"] - 2


# Test get progress density on the article
def test_get_progress_density(setup):
    client, _, project = setup
    # upload dataset
    au.upload_data_to_project(client, project, data=UPLOAD_DATA[0])
    # request progress density
    status_code, data = au.get_project_progress_density(client, project)
    assert status_code == 200
    assert isinstance(data, dict)
    assert isinstance(data["relevant"], list)
    assert isinstance(data["irrelevant"], list)


# Test progress recall
def test_get_progress_recall(setup):
    client, _, project = setup
    # upload dataset
    au.upload_data_to_project(client, project, data=UPLOAD_DATA[0])
    # get recall
    status_code, data = au.get_project_progress_recall(client, project)
    assert status_code == 200
    assert isinstance(data, dict)
    assert isinstance(data["asreview"], list)
    assert isinstance(data["random"], list)


# Test retrieve documents in order to review
def test_retrieve_document_for_review(setup):
    client, _, project = setup
    # start the show
    au.upload_label_set_and_start_model(client, project, UPLOAD_DATA[0])
    # get a document
    status_code, data = au.get_project_current_document(client, project)
    assert status_code == 200
    assert isinstance(data, dict)
    assert not data["pool_empty"]
    assert isinstance(data["result"], dict)
    assert isinstance(data["result"]["doc_id"], int)


# Test label a document after the model has been started
def test_label_a_document_with_running_model(setup):
    client, _, project = setup
    # start the show
    au.upload_label_set_and_start_model(client, project, UPLOAD_DATA[0])
    # get a document
    _, data = au.get_project_current_document(client, project)
    # get id
    doc_id = data["result"]["doc_id"]
    # label it
    status_code, data = au.label_project_record(
        client, project, doc_id, label=1, prior=0, note="note"
    )
    assert status_code == 200
    assert data["success"]
    time.sleep(10)


# Test update label of a document after the model has been started
def test_update_label_of_document_with_running_model(setup):
    client, _, project = setup
    # start the show
    au.upload_label_set_and_start_model(client, project, UPLOAD_DATA[0])
    # get a document
    _, data = au.get_project_current_document(client, project)
    # get id
    doc_id = data["result"]["doc_id"]
    # label it
    au.label_project_record(client, project, doc_id, label=1, prior=0, note="note")
    # change label
    status_code, data = au.update_label_project_record(
        client, project, doc_id, label=0, prior=0, note="changed note"
    )
    assert status_code == 200
    assert data["success"]
    time.sleep(10)


# Test deleting a project
def test_delete_project(setup):
    client, _, project = setup
    # delete project
    status_code, data = au.delete_project(client, project)
    assert status_code == 200
    assert data["success"]


@pytest.mark.parametrize(
    "api_call",
    [
        au.get_all_projects,
        au.create_project,
        au.create_project_from_dict,
        au.update_project,
        au.upgrade_project,
        au.get_project_stats,
        au.get_demo_data,
        au.upload_data_to_project,
        au.get_project_data,
        au.get_project_dataset_writer,
        au.search_project_data,
        au.get_prior_random_project_data,
        au.label_project_record,
        au.update_label_project_record,
        au.get_labeled_project_data,
        au.get_labeled_project_data_stats,
        au.get_project_algorithms_options,
        au.set_project_algorithms,
        au.get_project_algorithms,
        au.start_project_algorithms,
        au.get_project_status,
        au.set_project_status,
        au.export_project_dataset,
        au.export_project,
        au.get_project_progress,
        au.get_project_progress_density,
        au.get_project_progress_recall,
        au.get_project_current_document,
        au.delete_project,
    ],
)
def test_unauthorized_use_of_api_calls(setup, api_call):
    client, user, project = setup
    if current_app_is_authenticated():
        # signout the client
        au.signout_user(client)
        # inspect function
        sig = inspect.signature(api_call)
        # form parameters
        parms = []
        for par in sig.parameters.keys():
            annotation = sig.parameters[par].annotation
            if annotation == FlaskClient:
                parms.append(client)
            elif annotation == Union[Project, ASReviewProject]:
                parms.append(project)
            elif annotation == int:
                parms.append(1)
            elif annotation == str:
                parms.append("abc")
            elif annotation == dict:
                parms.append({})

        # make the api call
        status_code, data = api_call(*parms)
        assert status_code == 401
        assert data["message"] == "Login required."
    else:
        # no asserts in an unauthenticated app
        pass



# Copyright 2019-2022 The ASReview Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
from math import log

import numpy as np

try:
    import tensorflow as tf
    from tensorflow.keras.preprocessing.text import text_to_word_sequence
except ImportError:
    TF_AVAILABLE = False
else:
    TF_AVAILABLE = True
    try:
        tf.logging.set_verbosity(tf.logging.ERROR)
    except AttributeError:
        logging.getLogger("tensorflow").setLevel(logging.ERROR)


from asreview.models.feature_extraction.base import BaseFeatureExtraction
from asreview.models.feature_extraction.embedding_lstm import load_embedding
from asreview.utils import get_random_state


def _check_tensorflow():
    if not TF_AVAILABLE:
        raise ImportError("Install tensorflow package to use" " Embedding IDF.")


class EmbeddingIdf(BaseFeatureExtraction):
    """Embedding IDF feature extraction technique (``embedding-idf``).

    This model averages the weighted word vectors of all the words in the
    text, in order to get a single feature vector for each text. The weights
    are provided by the inverse document frequencies.

    .. note::

        This feature extraction technique requires ``tensorflow`` to be
        installed. Use ``pip install tensorflow`` or install all optional
        ASReview dependencies with ``pip install asreview[all]``

    Arguments
    ---------
    embedding_fp: str
        Path to embedding.

    """

    name = "embedding-idf"
    label = "Embedding IDF"

    def __init__(self, *args, embedding_fp=None, random_state=None, **kwargs):
        """Initialize the Embedding-Idf model."""
        super(EmbeddingIdf, self).__init__(*args, **kwargs)
        self.embedding_fp = embedding_fp
        self.embedding = None
        self._random_state = get_random_state(random_state)

    def transform(self, texts):
        # check is tensorflow is available
        _check_tensorflow()

        if self.embedding is None:
            if self.embedding_fp is None:
                raise ValueError("Error: need embedding to train Embeddingdf model.")
            self.embedding = load_embedding(self.embedding_fp, n_jobs=-1)

        text_counts = _get_freq_dict(texts)
        idf = _get_idf(text_counts)
        X = _get_X_from_dict(text_counts, idf, self.embedding, self._random_state)
        return X


def _get_freq_dict(all_text):
    text_dicts = []
    for text in all_text:
        cur_dict = {}
        word_sequence = text_to_word_sequence(text)
        for word in word_sequence:
            if word in cur_dict:
                cur_dict[word] += 1
            else:
                cur_dict[word] = 1
        text_dicts.append(cur_dict)
    return text_dicts


def _get_idf(text_dicts):
    all_count = {}
    for text in text_dicts:
        for word in text:
            if word in all_count:
                all_count[word] += 1
            else:
                all_count[word] = 1

    idf = {}
    for word in all_count:
        idf[word] = log(len(text_dicts) / all_count[word])
    return idf


def _get_X_from_dict(text_dicts, idf, embedding, random_state):
    n_vec = len(embedding[list(embedding.keys())[0]])
    X = np.zeros((len(text_dicts), n_vec))
    for i, text in enumerate(text_dicts):
        text_vec = None
        for word in text:
            cur_count = text[word]
            cur_idf = idf[word]
            cur_vec = embedding.get(word, None)
            if cur_vec is None:
                continue
            if text_vec is None:
                text_vec = cur_vec * cur_idf * cur_count
            else:
                text_vec += cur_vec * cur_idf * cur_count
        if text_vec is None:
            text_vec = random_state.random(n_vec)

        text_norm = np.linalg.norm(text_vec)
        if abs(text_norm) > 1e-7:
            text_vec /= np.linalg.norm(text_vec)
        X[i] = text_vec
    return X



# Copyright 2019-2022 The ASReview Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from asreview.utils import _entry_points


def list_query_strategies():
    """List available query strategy classes.

    This excludes all possible mixed query strategies.

    Returns
    -------
    list
        Classes of available query strategies in alphabetical order.
    """
    return [e.load() for e in _entry_points(group="asreview.models.query")]


def get_query_class(name):
    """Get class of query strategy from its name.

    Arguments
    ---------
    name: str
        Name of the query strategy, e.g. 'max', 'uncertainty', 'random.
        A special mixed query strategy is als possible. The mix is denoted
        by an underscore: 'max_random' or 'max_uncertainty'.

    Returns
    -------
    class
        Class corresponding to the name name.
    """

    return _entry_points(group="asreview.models.query")[name].load()


def get_query_model(name, *args, random_state=None, **kwargs):
    """Get an instance of the query strategy.

    Arguments
    ---------
    name: str
        Name of the query strategy.
    *args:
        Arguments for the model.
    **kwargs:
        Keyword arguments for the model.

    Returns
    -------
    asreview.query.base.BaseQueryModel
        Initialized instance of query strategy.
    """
    query_class = get_query_class(name)
    try:
        return query_class(*args, random_state=random_state, **kwargs)
    except TypeError:
        return query_class(*args, **kwargs)



# Copyright 2019-2022 The ASReview Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import argparse
import logging
import os
import socket
import webbrowser
from pathlib import Path
from threading import Timer

try:
    import tomllib
except ImportError:
    import tomli as tomllib

from flask import Flask
from flask import send_from_directory
from flask.json import jsonify
from flask.templating import render_template
from flask_cors import CORS
from flask_login import LoginManager
from gevent.pywsgi import WSGIServer
from werkzeug.exceptions import InternalServerError

from asreview import __version__ as asreview_version
from asreview._deprecated import DeprecateAction
from asreview._deprecated import mark_deprecated_help_strings
from asreview.project import ASReviewProject
from asreview.project import get_project_path
from asreview.project import get_projects
from asreview.utils import asreview_path
from asreview.webapp import DB
from asreview.webapp.api import auth
from asreview.webapp.api import projects
from asreview.webapp.api import team
from asreview.webapp.authentication.models import User
from asreview.webapp.authentication.oauth_handler import OAuthHandler

# Host name
HOST_NAME = os.getenv("ASREVIEW_HOST")
if HOST_NAME is None:
    HOST_NAME = "localhost"
# Default Port number
PORT_NUMBER = 5000

# set logging level
if (
    os.environ.get("FLASK_DEBUG", "") == "1"
    or os.environ.get("DEBUG", "") == "1"
    or os.environ.get("FLASK_ENV", "") == "development"
):
    logging.basicConfig(level=logging.DEBUG)
else:
    logging.basicConfig(level=logging.INFO)


def _url(host, port, protocol):
    """Create url from host and port."""
    return f"{protocol}{host}:{port}/"


def _check_port_in_use(host, port):
    """Check if port is already in use.

    Arguments
    ---------
    host: str
        The current host.
    port: int
        The host port to be checked.

    Returns
    -------
    bool:
        True if port is in use, false otherwise.
    """
    logging.info(f"Checking if host and port are available :: {host}:{port}")
    host = host.replace("https://", "").replace("http://", "")
    with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
        return s.connect_ex((host, port)) == 0


def _open_browser(host, port, protocol, no_browser):
    """Open ASReview in browser if flag is set.

    Otherwise, it displays an alert to copy and paste the url
    at which ASReview is currently served.
    """
    if no_browser:
        print(
            "\nTo access ASReview LAB, copy and paste "
            "this url in a browser "
            f"{_url(host, port, protocol)}\n"
        )
        return

    start_url = _url(host, port, protocol)
    Timer(1, lambda: webbrowser.open_new(start_url)).start()
    print(
        f"Start browser at {start_url}"
        "\n\n\n\nIf your browser doesn't open. "
        f"Please navigate to '{start_url}'\n\n\n\n"
    )


def _lab_parser():
    # parse arguments if available
    parser = argparse.ArgumentParser(
        prog="lab",
        description="""ASReview LAB - Active learning for Systematic Reviews.""",  # noqa
        formatter_class=argparse.RawTextHelpFormatter,
    )

    parser.add_argument(
        "--clean-project",
        dest="clean_project",
        default=None,
        type=str,
        help="Safe cleanup of temporary files in project.",
    )

    parser.add_argument(
        "--clean-all-projects",
        dest="clean_all_projects",
        default=None,
        action="store_true",
        help="Safe cleanup of temporary files in all projects.",
    )

    parser.add_argument(
        "--ip",
        default=HOST_NAME,
        type=str,
        action=DeprecateAction,
        help="The IP address the server will listen on. Use the --host argument.",
    )

    parser.add_argument(
        "--host",
        default=HOST_NAME,
        type=str,
        help="The IP address the server will listen on.",
    )

    parser.add_argument(
        "--port",
        default=PORT_NUMBER,
        type=int,
        help="The port the server will listen on.",
    )

    parser.add_argument(
        "--enable-auth",
        dest="enable_authentication",
        action="store_true",
        help="Enable authentication.",
    )

    parser.add_argument(
        "--auth-database-uri",
        default=None,
        type=str,
        help="URI of authentication database.",
    )

    parser.add_argument(
        "--secret-key",
        default=None,
        type=str,
        help="Secret key for authentication.",
    )

    parser.add_argument(
        "--salt",
        default=None,
        type=str,
        help="When using authentication, a salt code is needed" "for hasing passwords.",
    )

    parser.add_argument(
        "--flask-configfile",
        default="",
        type=str,
        help="Full path to a TOML file containing Flask parameters"
        "for authentication.",
    )

    parser.add_argument(
        "--no-browser",
        dest="no_browser",
        action="store_true",
        help="Do not open ASReview LAB in a browser after startup.",
    )

    parser.add_argument(
        "--port-retries",
        dest="port_retries",
        default=50,
        type=int,
        help="The number of additional ports to try if the"
        "specified port is not available.",
    )

    parser.add_argument(
        "--certfile",
        default="",
        type=str,
        help="The full path to an SSL/TLS certificate file.",
    )

    parser.add_argument(
        "--keyfile",
        default="",
        type=str,
        help="The full path to a private key file for usage with SSL/TLS.",
    )

    parser.add_argument(
        "--config_file",
        type=str,
        default=None,
        help="Deprecated, see subcommand simulate.",
        action=DeprecateAction,
    )

    parser.add_argument(
        "--seed",
        default=None,
        type=int,
        help="Deprecated, see subcommand simulate.",
        action=DeprecateAction,
    )

    parser.add_argument(
        "--embedding",
        type=str,
        default=None,
        dest="embedding_fp",
        help="File path of embedding matrix. Required for LSTM models.",
    )
    return parser


def create_app(**kwargs):
    app = Flask(
        __name__,
        instance_relative_config=True,
        static_folder="build/static",
        template_folder="build",
    )

    # Get the ASReview arguments.
    app.config["asr_kwargs"] = kwargs
    app.config["AUTHENTICATION_ENABLED"] = kwargs.get("enable_authentication", False)
    app.config["SECRET_KEY"] = kwargs.get("secret_key", False)
    app.config["SECURITY_PASSWORD_SALT"] = kwargs.get("salt", False)
    app.config["PORT"] = kwargs.get("port")
    app.config["HOST"] = kwargs.get("host")

    # Read config parameters if possible, this overrides
    # the previous assignments. Flask config parameters may come
    # as an environment var or from an argument. Argument
    # takes precedence.
    config_from_env = os.environ.get("FLASK_CONFIGFILE", "").strip()
    config_from_arg = kwargs.get("flask_configfile", "").strip()
    config_file_path = config_from_arg or config_from_env

    # Use absolute path, because otherwise it is relative to the config root.
    if config_file_path != "":
        config_file_path = Path(config_file_path)
        if config_file_path.suffix == ".toml":
            app.config.from_file(
                config_file_path.absolute(),
                load=tomllib.load,
                text=False
            )
        else:
            raise ValueError("'flask_configfile' should have a .toml extension")

    # If the frontend runs on a different port, or even on a different
    # URL, then allowed-origins must be set to avoid CORS issues. You can
    # set the allowed-origins in the config file. In the previous lines
    # the config file has been read.
    # If the allowed-origins are not set by now, they are set to
    # False, which will bypass setting any CORS parameters!
    if not app.config.get("ALLOWED_ORIGINS", False):
        app.config["ALLOWED_ORIGINS"] = False

    if not app.config["ALLOWED_ORIGINS"] and os.environ.get("FLASK_DEBUG", "") == "1":
        app.config["ALLOWED_ORIGINS"] = [f"http://{app.config['HOST']}:3000"]

    # set env (test / development / production) according to
    # Flask 2.2 specs (ENV is deprecated)
    if app.config.get("TESTING", None) is True:
        env = "test"
    elif app.config.get("DEBUG", None) is True:
        env = "development"
    else:
        env = "production"

    # config JSON Web Tokens
    login_manager = LoginManager(app)
    login_manager.init_app(app)
    login_manager.session_protection = "strong"

    if app.config["AUTHENTICATION_ENABLED"] is False:
        # This ensures the app handles the anonymous user
        # when authentication is disabled and there is no
        # configuration file
        app.config["SECRET_KEY"] = ""

        # This is necessary to pass the test_webapp.py tests
        @login_manager.user_loader
        def load_user(user_id):
            return False

    # setup all database/authentication related resources,
    # only do this when AUTHENTICATION_ENABLED is explicitly True
    elif app.config["AUTHENTICATION_ENABLED"] is True:
        # Register a callback function for current_user.
        @login_manager.user_loader
        def load_user(user_id):
            return User.query.get(int(user_id))

        # In this code-block we make sure certain authentication-related
        # config parameters are set.
        # TODO: should I raise a custom Exception, like MissingParameterError?
        if not app.config.get("SECRET_KEY", False):
            raise ValueError(
                "Please start an authenticated app with a "
                + "secret key parameter (SECRET_KEY)"
            )

        if not app.config.get("SECURITY_PASSWORD_SALT", False):
            raise ValueError(
                "Please start an authenticated app with a "
                + "security password salt (SECURITY_PASSWORD_SALT)"
            )

        if app.config.get("EMAIL_VERIFICATION", False) and not app.config.get(
            "EMAIL_CONFIG", False
        ):
            raise ValueError(
                "Missing email configuration to facilitate email verification"
            )

        # set email config for Flask-Mail
        conf = app.config.get("EMAIL_CONFIG", {})
        app.config["MAIL_SERVER"] = conf.get("SERVER")
        app.config["MAIL_PORT"] = conf.get("PORT", 465)
        app.config["MAIL_USERNAME"] = conf.get("USERNAME")
        app.config["MAIL_PASSWORD"] = conf.get("PASSWORD")
        app.config["MAIL_USE_TLS"] = conf.get("USE_TLS", False)
        app.config["MAIL_USE_SSL"] = conf.get("USE_SSL", False)
        app.config["MAIL_REPLY_ADDRESS"] = conf.get("REPLY_ADDRESS")

        # We must be sure we have a SQLAlchemy database URI. At this
        # stage the TOML file has been read. See if we haven't found
        # such a URI.
        if not app.config.get("SQLALCHEMY_DATABASE_URI", False):
            # there is no configuration, check CLI parameters
            cli_database_uri = (kwargs.get("auth_database_uri") or "").strip()

            # if we still haven't found a database URI, create a sqlite3 database
            if cli_database_uri != "":
                app.config["SQLALCHEMY_DATABASE_URI"] = cli_database_uri
            else:
                # create default path
                uri = os.path.join(asreview_path(), f"asreview.{env}.sqlite")
                app.config["SQLALCHEMY_DATABASE_URI"] = f"sqlite:///{uri}"

        # initialize app for SQLAlchemy
        DB.init_app(app)

        with app.app_context():
            # create tables in case they don't exist
            DB.create_all()

        # store oauth config in oauth handler
        if bool(app.config.get("OAUTH", False)):
            app.config["OAUTH"] = OAuthHandler(app.config["OAUTH"])

    # Ensure the instance folder exists.
    try:
        os.makedirs(app.instance_path)
    except OSError:
        pass

    # We only need CORS if they are necessary: when the frontend is
    # running on a different port, or even url, we need to set the
    # allowed origins to avoid CORS problems. The allowed-origins
    # can be set in the config file.
    if app.config.get("ALLOWED_ORIGINS", False):
        CORS(app, origins=app.config.get("ALLOWED_ORIGINS"), supports_credentials=True)

    with app.app_context():
        app.register_blueprint(projects.bp)
        app.register_blueprint(auth.bp)
        app.register_blueprint(team.bp)

    @app.errorhandler(InternalServerError)
    def error_500(e):
        original = getattr(e, "original_exception", None)

        if original is None:
            # direct 500 error, such as abort(500)
            logging.error(e)
            return jsonify(message="Whoops, something went wrong."), 500

        # wrapped unhandled error
        logging.error(e.original_exception)
        return jsonify(message=str(e.original_exception)), 500

    @app.route("/", methods=["GET"])
    @app.route("/confirm_account", methods=["GET"])
    @app.route("/oauth_callback", methods=["GET"])
    @app.route("/projects/", methods=["GET"])
    @app.route("/projects/<project_id>/", methods=["GET"])
    @app.route("/projects/<project_id>/<tab>/", methods=["GET"])
    @app.route("/reset_password", methods=["GET"])
    def index(**kwargs):
        return render_template("index.html")

    @app.route("/favicon.ico")
    def send_favicon():
        return send_from_directory(
            "build", "favicon.ico", mimetype="image/vnd.microsoft.icon"
        )

    @app.route("/boot", methods=["GET"])
    def api_boot():
        """Get the boot info."""
        if app.config.get("DEBUG", None) is True:
            status = "development"
        else:
            status = "asreview"

        # the big one
        authenticated = app.config.get("AUTHENTICATION_ENABLED", False)

        response = {
            "status": status,
            "authentication": authenticated,
            "version": asreview_version,
        }

        # if we do authentication we have a lot of extra parameters
        if authenticated:
            # if recaptcha config is provided for account creation
            if app.config.get("RE_CAPTCHA_V3", False):
                response["recaptchav3_key"] = app.config["RE_CAPTCHA_V3"].get(
                    "KEY", False
                )

            # check if users can create accounts
            response["allow_account_creation"] = app.config.get(
                "ALLOW_ACCOUNT_CREATION", False
            )

            response["allow_teams"] = app.config.get("ALLOW_TEAMS", False)

            # check if we are doing email verification
            response["email_verification"] = bool(
                app.config.get("EMAIL_VERIFICATION", False)
            )

            # check if there is an email server setup (forgot password)
            response["email_config"] = bool(app.config.get("EMAIL_CONFIG", False))

            # if oauth config is provided
            if isinstance(app.config.get("OAUTH", False), OAuthHandler):
                params = app.config.get("OAUTH").front_end_params()
                # and there something in it, just to be sure
                if params:
                    response["oauth"] = params

        return jsonify(response)

    return app


def main(argv):
    parser = _lab_parser()
    mark_deprecated_help_strings(parser)
    args = parser.parse_args(argv)

    app = create_app(**vars(args))
    app.config["PROPAGATE_EXCEPTIONS"] = False

    # ssl certificate, key and protocol
    certfile = args.certfile
    keyfile = args.keyfile
    ssl_context = None
    if certfile and keyfile:
        protocol = "https://"
        ssl_context = (certfile, keyfile)
    else:
        protocol = "http://"

    # clean all projects
    # TODO@{Casper}: this needs a little bit
    # of work, we need to access all sub-folders
    if args.clean_all_projects:
        print("Cleaning all project files.")
        for project in get_projects():
            project.clean_tmp_files()
        print("Done")
        return

    # clean project by project_id
    # TODO@{Casper}: cleaning without a user context
    # is meaningless -> I think we should remove this
    # option
    if args.clean_project is not None:
        print(f"Cleaning project file '{args.clean_project}'.")
        ASReviewProject(get_project_path(args.clean_project)).clean_tmp_files()
        print("Done")
        return

    flask_dev = app.config.get("DEBUG", False)

    host = app.config.get("HOST")
    port = app.config.get("PORT")

    port_retries = args.port_retries
    # if port is already taken find another one
    if not flask_dev:
        original_port = port
        while _check_port_in_use(host, port) is True:
            old_port = port
            port = int(port) + 1
            if port - original_port >= port_retries:
                raise ConnectionError(
                    "Could not find an available port \n"
                    "to launch ASReview LAB. Last port \n"
                    f"was {str(port)}"
                )
            print(f"Port {old_port} is in use.\n* Trying to start at {port}")

    # open webbrowser if not in flask development mode
    if flask_dev is False:
        _open_browser(host, port, protocol, args.no_browser)

    # run app in flask mode only if we run in development mode
    if flask_dev is True:
        app.run(host=host, port=port, ssl_context=ssl_context)
    else:
        ssl_args = {"keyfile": keyfile, "certfile": certfile} if ssl_context else {}
        server = WSGIServer((host, port), app, **ssl_args)
        server.serve_forever()



from pathlib import Path
from sqlite3 import OperationalError

import pandas as pd
import pytest
from scipy.sparse import csr_matrix

from asreview import ASReviewData
from asreview.project import ASReviewProject
from asreview.project import ProjectExistsError
from asreview.project import open_state
from asreview.settings import ASReviewSettings
from asreview.state import SQLiteState
from asreview.state.errors import StateNotFoundError
from asreview.state.sqlstate import RESULTS_TABLE_COLUMNS

TEST_LABELS = [1, 0, 0, 1, 1, 1, 0, 1, 1, 1]
TEST_INDICES = [16, 346, 509, 27, 11, 555, 554, 680, 264, 309]
TEST_RECORD_IDS = [17, 347, 510, 28, 12, 556, 555, 681, 265, 310]
TEST_RECORD_TABLE = list(range(851))
TEST_CLASSIFIERS = [None, None, None, None, "nb", "nb", "nb", "nb", "nb", "nb"]
TEST_QUERY_STRATEGIES = [
    "prior",
    "prior",
    "prior",
    "prior",
    "max",
    "max",
    "max",
    "max",
    "max",
    "max",
]
TEST_BALANCE_STRATEGIES = [
    None,
    None,
    None,
    None,
    "double",
    "double",
    "double",
    "double",
    "double",
    "double",
]
TEST_FEATURE_EXTRACTION = [
    None,
    None,
    None,
    None,
    "tfidf",
    "tfidf",
    "tfidf",
    "tfidf",
    "tfidf",
    "tfidf",
]
TEST_TRAINING_SETS = [-1, -1, -1, -1, 4, 5, 6, 7, 8, 9]
TEST_NOTES = [
    None,
    None,
    None,
    "random text",
    "another random text",
    None,
    None,
    "A final random text",
    None,
    None,
]

TEST_N_PRIORS = 4
TEST_N_MODELS = 7

TEST_STATE_FP = Path("tests", "asreview_files", "test_state_example_converted.asreview")
TEST_WITH_TIMES_FP = Path(
    "tests", "asreview_files", "test_state_example_with_times.asreview"
)
TEST_LABELING_TIMES = [
    "2021-09-30 17:54:07.569255",
    "2021-09-30 17:54:07.569255",
    "2021-09-30 17:54:28.860270",
    "2021-09-30 17:54:28.860270",
    "2021-09-30 17:54:28.860270",
    "2021-09-30 17:54:31.689389",
    "2021-09-30 17:54:33.505257",
    "2021-09-30 17:54:35.842416",
    "2021-09-30 17:54:38.245108",
]

TEST_FIRST_PROBS = [
    0.7107394917661797,
    0.7291694332065035,
    0.732624685298732,
    0.7017866934752249,
    0.7275304788204621,
    0.7126109527686055,
    0.7246720268636593,
    0.7040374218528891,
    0.7095665447517838,
    0.7021937381372063,
]
TEST_LAST_PROBS = [
    0.7116408177006979,
    0.7119557616570122,
    0.71780127925996,
    0.7127075014419986,
    0.7085644453092131,
    0.7067520535764322,
    0.7103161247883791,
    0.7192568428839242,
    0.7118104532649111,
    0.7150387267232563,
]
TEST_POOL_START = [157, 301, 536, 567, 416, 171, 659, 335, 329, 428]


def test_init_project_folder(tmpdir):
    project_path = Path(tmpdir, "test.asreview")
    project = ASReviewProject.create(project_path)

    assert Path(project_path, "project.json").is_file()
    assert Path(project_path, "data").is_dir()
    assert Path(project_path, "feature_matrices").is_dir()
    assert Path(project_path, "reviews").is_dir()

    assert project.config["id"] == "test"


def test_init_project_already_exists(tmpdir):
    project_path = Path(tmpdir, "test.asreview")
    ASReviewProject.create(project_path)
    with pytest.raises(ProjectExistsError):
        ASReviewProject.create(project_path)


def test_invalid_project_folder():
    with pytest.raises(StateNotFoundError):
        with open_state("this_is_not_a_project") as state:  # noqa
            pass


def test_state_not_found(tmpdir):
    project_path = Path(tmpdir, "test.asreview")
    ASReviewProject.create(project_path)
    with pytest.raises(StateNotFoundError):
        with open_state(project_path) as state:  # noqa
            pass


def test_read_basic_state():
    with open_state(TEST_STATE_FP) as state:
        assert isinstance(state, SQLiteState)


def test_version_number_state():
    with open_state(TEST_STATE_FP) as state:
        assert state.version[0] == "1"


def test_write_while_read_only_state():
    with open_state(TEST_STATE_FP, read_only=True) as state:
        with pytest.raises(OperationalError):
            state.add_last_probabilities([1.0] * len(TEST_RECORD_TABLE))


def test_print_state():
    with open_state(TEST_STATE_FP, read_only=True) as state:
        print(state)


def test_settings_state():
    with open_state(TEST_STATE_FP) as state:
        assert isinstance(state.settings, ASReviewSettings)


def test_n_records_labeled():
    with open_state(TEST_STATE_FP) as state:
        assert state.n_records_labeled == len(TEST_LABELS)


def test_n_priors():
    with open_state(TEST_STATE_FP) as state:
        assert state.n_priors == TEST_N_PRIORS


def test_create_new_state_file(tmpdir):
    project_path = Path(tmpdir, "test.asreview")
    ASReviewProject.create(project_path)
    with open_state(project_path, read_only=False) as state:
        state._is_valid_state()


def test_get_dataset():
    with open_state(TEST_STATE_FP) as state:
        assert isinstance(state.get_dataset(["query_strategy"]), pd.DataFrame)
        assert isinstance(state.get_dataset(), pd.DataFrame)

        # Try getting a specific column.
        assert (
            state.get_dataset(["record_id"])["record_id"].to_list() == TEST_RECORD_IDS
        )
        assert (
            state.get_dataset(["feature_extraction"])["feature_extraction"].to_list()
            == TEST_FEATURE_EXTRACTION
        )
        # Try getting all columns and that picking the right column.
        assert (
            state.get_dataset()["balance_strategy"].to_list() == TEST_BALANCE_STRATEGIES
        )
        # Try getting a specific column with column name as string, instead of
        # list containing column name.
        assert (
            state.get_dataset("training_set")["training_set"].to_list()
            == TEST_TRAINING_SETS
        )


def test_get_dataset_drop_prior():
    with open_state(TEST_STATE_FP) as state:
        assert (
            len(state.get_dataset(priors=False)) == len(TEST_RECORD_IDS) - TEST_N_PRIORS
        )
        assert (state.get_dataset(priors=False)["query_strategy"] != "prior").all()
        assert "query_strategy" in state.get_dataset(priors=False).columns
        assert "query_strategy" not in state.get_dataset("label", priors=False)


def test_get_dataset_drop_pending(tmpdir):
    record_table = range(1, 11)
    test_ranking = range(10, 0, -1)
    project_path = Path(tmpdir, "test.asreview")
    ASReviewProject.create(project_path)
    with open_state(project_path, read_only=False) as state:
        state.add_record_table(record_table)
        state.add_last_ranking(test_ranking, "nb", "max", "double", "tfidf", 4)
        state.add_labeling_data([4, 5, 6], [1, 0, 1], prior=True)
        state.query_top_ranked(3)

        assert "label" in state.get_dataset(pending=False).columns
        assert "label" not in state.get_dataset("balance_strategy", pending=False)
        assert len(state.get_dataset(pending=False)) == 3
        assert state.get_dataset(pending=False)["label"].notna().all()


def test_get_data_by_query_number():
    with open_state(TEST_STATE_FP) as state:
        query = state.get_data_by_query_number(0)
        assert list(query.columns) == RESULTS_TABLE_COLUMNS
        assert (
            query["balance_strategy"].tolist()
            == TEST_BALANCE_STRATEGIES[:TEST_N_PRIORS]
        )
        assert query["classifier"].tolist() == TEST_CLASSIFIERS[:TEST_N_PRIORS]

        for query_num in [1, 3, 5]:
            query_idx = query_num + TEST_N_PRIORS - 1
            query = state.get_data_by_query_number(query_num)
            assert isinstance(query, pd.DataFrame)
            assert (
                query["feature_extraction"].to_list()[0]
                == TEST_FEATURE_EXTRACTION[query_idx]
            )
            assert query["label"].to_list()[0] == TEST_LABELS[query_idx]
            assert query["record_id"].to_list()[0] == TEST_RECORD_IDS[query_idx]

        columns = RESULTS_TABLE_COLUMNS[2:5]
        query = state.get_data_by_query_number(4, columns)
        assert list(query.columns) == columns


def test_get_data_by_record_id():
    with open_state(TEST_STATE_FP) as state:
        for idx in [2, 6, 8]:
            record_id = TEST_RECORD_IDS[idx]
            query = state.get_data_by_record_id(record_id)
            assert isinstance(query, pd.DataFrame)
            assert query["training_set"].to_list()[0] == TEST_TRAINING_SETS[idx]
            assert query["record_id"].to_list()[0] == TEST_RECORD_IDS[idx]


def test_get_query_strategies():
    with open_state(TEST_STATE_FP) as state:
        assert isinstance(state.get_query_strategies(), pd.Series)
        assert state.get_query_strategies().to_list() == TEST_QUERY_STRATEGIES


def test_get_classifiers():
    with open_state(TEST_STATE_FP) as state:
        assert isinstance(state.get_classifiers(), pd.Series)
        assert state.get_classifiers().to_list() == TEST_CLASSIFIERS


def test_get_training_sets():
    with open_state(TEST_STATE_FP) as state:
        assert isinstance(state.get_training_sets(), pd.Series)
        assert all(state.get_training_sets() == TEST_TRAINING_SETS)


def test_get_order_of_labeling():
    with open_state(TEST_STATE_FP) as state:
        assert isinstance(state.get_order_of_labeling(), pd.Series)
        assert all(state.get_order_of_labeling() == TEST_RECORD_IDS)


def test_get_labels():
    with open_state(TEST_STATE_FP) as state:
        assert isinstance(state.get_labels(), pd.Series)
        assert all(state.get_labels() == TEST_LABELS)


def test_get_labels_no_priors():
    with open_state(TEST_STATE_FP) as state:
        labels = state.get_labels(priors=False)
        assert isinstance(labels, pd.Series)
        assert all(labels == TEST_LABELS[4:])


def test_get_labeling_times():
    with open_state(TEST_WITH_TIMES_FP) as state:
        assert isinstance(state.get_labeling_times(), pd.Series)
        assert all(state.get_labeling_times() == TEST_LABELING_TIMES)


def test_create_empty_state(tmpdir):
    project_path = Path(tmpdir, "test.asreview")
    ASReviewProject.create(project_path)
    with open_state(project_path, read_only=False) as state:
        assert state.is_empty()


def test_get_feature_matrix():
    project = ASReviewProject(TEST_STATE_FP)

    assert len(project.feature_matrices) == 1

    feature_matrix = project.get_feature_matrix(project.feature_matrices[0]["id"])
    assert isinstance(feature_matrix, csr_matrix)


def test_get_record_table():
    with open_state(TEST_STATE_FP) as state:
        record_table = state.get_record_table()
        assert isinstance(record_table, pd.Series)
        assert record_table.name == "record_id"
        assert record_table.to_list() == TEST_RECORD_TABLE


def test_record_table(tmpdir):
    data_fp = Path("tests", "demo_data", "record_id.csv")
    as_data = ASReviewData.from_file(data_fp)

    project_path = Path(tmpdir, "test.asreview")
    ASReviewProject.create(project_path)

    with open_state(project_path, read_only=False) as state:
        state.add_record_table(as_data.record_ids)
        assert state.get_record_table().to_list() == list(range(len(as_data)))


def test_get_last_probabilities():
    with open_state(TEST_STATE_FP) as state:
        probabilities = state.get_last_probabilities()
        assert isinstance(probabilities, pd.Series)
        assert probabilities.name == "proba"
        assert probabilities.to_list()[:10] == TEST_FIRST_PROBS
        assert probabilities.to_list()[-10:] == TEST_LAST_PROBS


def test_add_last_probabilities_fail():
    with open_state(TEST_STATE_FP) as state:
        with pytest.raises(ValueError):
            state.add_last_probabilities([1.0, 2.0, 3.0])


def test_add_last_probabilities(tmpdir):
    project_path = Path(tmpdir, "test.asreview")
    ASReviewProject.create(project_path)
    probabilities = [float(num) for num in range(50)]
    with open_state(project_path, read_only=False) as state:
        state.add_last_probabilities(probabilities)
        state_probabilities = state.get_last_probabilities().to_list()
        assert state_probabilities == probabilities


def test_move_ranking_data_to_results(tmpdir):
    project_path = Path(tmpdir, "test.asreview")
    ASReviewProject.create(project_path)
    with open_state(project_path, read_only=False) as state:
        state.add_record_table(TEST_RECORD_TABLE)
        state.add_last_ranking(
            range(1, len(TEST_RECORD_TABLE) + 1), "nb", "max", "double", "tfidf", 4
        )
        state._move_ranking_data_to_results([4, 6, 5, 7])

        data = state.get_dataset(pending=True)
        assert data["record_id"].to_list() == [4, 6, 5, 7]
        assert data["label"].to_list() == [None] * 4
        assert data["classifier"].to_list() == ["nb"] * 4


def test_query_top_ranked(tmpdir):
    test_ranking = [2, 1, 0] + list(range(3, len(TEST_RECORD_TABLE)))
    project_path = Path(tmpdir, "test.asreview")
    ASReviewProject.create(project_path)
    with open_state(project_path, read_only=False) as state:
        state.add_record_table(TEST_RECORD_TABLE)
        state.add_last_ranking(test_ranking, "nb", "max", "double", "tfidf", 4)
        top_ranked = state.query_top_ranked(5)

        assert top_ranked == [2, 1, 0, 3, 4]
        data = state.get_dataset(pending=True)
        assert data["record_id"].to_list() == [2, 1, 0, 3, 4]
        assert data["classifier"].to_list() == ["nb"] * 5
        assert data["query_strategy"].to_list() == ["max"] * 5
        assert data["balance_strategy"].to_list() == ["double"] * 5
        assert data["feature_extraction"].to_list() == ["tfidf"] * 5
        assert data["training_set"].to_list() == [4] * 5


def test_add_labeling_data(tmpdir):
    test_ranking = list(range(len(TEST_RECORD_TABLE)))
    project_path = Path(tmpdir, "test.asreview")
    ASReviewProject.create(project_path)
    with open_state(project_path, read_only=False) as state:
        state.add_record_table(TEST_RECORD_TABLE)
        state.add_last_ranking(test_ranking, "nb", "max", "double", "tfidf", 4)
        for i in range(3):
            # Test without specifying notes.
            state.add_labeling_data([TEST_RECORD_IDS[i]], [TEST_LABELS[i]], prior=True)

        # Test with specifying notes and with larger batch.
        state.add_labeling_data(
            TEST_RECORD_IDS[3:6], TEST_LABELS[3:6], notes=TEST_NOTES[3:6], prior=True
        )

        data = state.get_dataset(pending=True)
        assert data["record_id"].to_list() == TEST_RECORD_IDS[:6]
        assert data["label"].to_list() == TEST_LABELS[:6]
        assert data["classifier"].to_list() == [None] * 6
        assert data["query_strategy"].to_list() == ["prior"] * 6
        assert data["balance_strategy"].to_list() == [None] * 6
        assert data["feature_extraction"].to_list() == [None] * 6
        assert data["training_set"].to_list() == [-1] * 6
        assert data["notes"].to_list() == TEST_NOTES[:6]

        state.query_top_ranked(3)
        data = state.get_dataset(pending=True)
        assert data["label"].to_list()[:6] == TEST_LABELS[:6]
        assert data["label"][6:].isna().all()
        assert data["record_id"].to_list() == TEST_RECORD_IDS[:6] + [0, 1, 2]

        state.add_labeling_data([1], [1])
        labels = state.get_labels(pending=True)
        assert labels.to_list()[:6] == TEST_LABELS[:6]
        assert labels[7] == 1

        state.add_labeling_data([0, 2], [0, 1], notes=["note0", "note2"])
        data = state.get_dataset(pending=True)
        assert data["label"].to_list() == TEST_LABELS[:6] + [0, 1, 1]
        assert data["notes"].to_list() == TEST_NOTES[:6] + ["note0", None, "note2"]


def test_pool_labeled_pending(tmpdir):
    record_table = range(1, 11)
    test_ranking = range(10, 0, -1)
    project_path = Path(tmpdir, "test.asreview")
    ASReviewProject.create(project_path)
    with open_state(project_path, read_only=False) as state:
        state.add_record_table(record_table)
        state.add_last_ranking(test_ranking, "nb", "max", "double", "tfidf", 4)
        state.add_labeling_data([4, 5, 6], [1, 0, 1], prior=True)
        state.query_top_ranked(3)

        pool, labeled, pending = state.get_pool_labeled_pending()
        assert isinstance(pool, pd.Series)
        assert isinstance(labeled, pd.DataFrame)
        assert isinstance(pending, pd.Series)

        assert pool.name == "record_id"
        assert pending.name == "record_id"
        assert list(labeled.columns) == ["record_id", "label"]

        assert pool.to_list() == [7, 3, 2, 1]
        assert labeled["record_id"].to_list() == [4, 5, 6]
        assert labeled["label"].to_list() == [1, 0, 1]
        assert pending.to_list() == [10, 9, 8]

        pool2 = state.get_pool()
        labeled2 = state.get_labeled()
        pending2 = state.get_pending()

        assert isinstance(pool2, pd.Series)
        assert isinstance(labeled2, pd.DataFrame)
        assert isinstance(pending2, pd.Series)

        assert pool2.name == "record_id"
        assert pending2.name == "record_id"
        assert list(labeled2.columns) == ["record_id", "label"]

        assert pool.to_list() == pool2.to_list()
        assert labeled["record_id"].to_list() == labeled2["record_id"].to_list()
        assert labeled["label"].to_list() == labeled2["label"].to_list()
        assert pending.to_list() == pending2.to_list()


def test_exist_new_labeled_records(tmpdir):
    record_table = range(1, 11)
    test_ranking = range(10, 0, -1)
    project_path = Path(tmpdir, "test.asreview")
    ASReviewProject.create(project_path)
    with open_state(project_path, read_only=False) as state:
        state.add_record_table(record_table)

        assert not state.exist_new_labeled_records
        state.add_labeling_data([4, 5, 6], [1, 0, 1], prior=True)
        assert state.exist_new_labeled_records
        state.add_last_ranking(test_ranking, "nb", "max", "double", "tfidf", 3)
        assert not state.exist_new_labeled_records
        state.query_top_ranked(3)
        assert not state.exist_new_labeled_records
        state.add_labeling_data([8, 9, 10], [1, 1, 1])
        assert state.exist_new_labeled_records


def test_add_note(tmpdir):
    project_path = Path(tmpdir, "test.asreview")
    ASReviewProject.create(project_path)
    with open_state(project_path, read_only=False) as state:
        state.add_record_table(TEST_RECORD_TABLE)
        state.add_labeling_data(
            TEST_RECORD_IDS[:3], TEST_LABELS[:3], TEST_NOTES[:3], prior=True
        )

        note = "An added note"
        record_id = TEST_RECORD_IDS[1]
        state.add_note(note, record_id)
        record_data = state.get_data_by_record_id(record_id)
        assert record_data["notes"][0] == note


def test_update_decision(tmpdir):
    project_path = Path(tmpdir, "test.asreview")
    ASReviewProject.create(project_path)
    with open_state(project_path, read_only=False) as state:
        state.add_record_table(TEST_RECORD_TABLE)
        state.add_labeling_data(TEST_RECORD_IDS[:3], TEST_LABELS[:3], prior=True)

        for i in range(3):
            state.update_decision(TEST_RECORD_IDS[i], 1 - TEST_LABELS[i])
            new_label = state.get_data_by_record_id(TEST_RECORD_IDS[i])["label"][0]
            assert new_label == 1 - TEST_LABELS[i]

        state.update_decision(TEST_RECORD_IDS[1], TEST_LABELS[1])
        new_label = state.get_data_by_record_id(TEST_RECORD_IDS[1])["label"][0]
        assert new_label == TEST_LABELS[1]

        change_table = state.get_decision_changes()
        changed_records = TEST_RECORD_IDS[:3] + [TEST_RECORD_IDS[1]]
        new_labels = [1 - x for x in TEST_LABELS[:3]] + [TEST_LABELS[1]]

        assert change_table["record_id"].to_list() == changed_records
        assert change_table["new_label"].to_list() == new_labels


def test_get_pool_labeled():
    with open_state(TEST_STATE_FP) as state:
        pool, labeled, _ = state.get_pool_labeled_pending()

    assert isinstance(pool, pd.Series)
    assert pool.name == "record_id"
    assert isinstance(labeled, pd.DataFrame)
    assert list(labeled.columns) == ["record_id", "label"]

    assert pool.to_list()[:10] == TEST_POOL_START
    assert labeled["record_id"].to_list() == TEST_RECORD_IDS
    assert labeled["label"].to_list() == TEST_LABELS


def test_last_ranking(tmpdir):
    project_path = Path(tmpdir, "test.asreview")
    ASReviewProject.create(project_path)

    record_ids = [1, 2, 3, 4, 5, 6]
    ranking = [1, 3, 4, 6, 2, 5]
    classifier = "nb"
    query_strategy = "max"
    balance_strategy = "double"
    feature_extraction = "tfidf"
    training_set = 2

    with open_state(project_path, read_only=False) as state:
        state.add_record_table(record_ids)
        state.add_last_ranking(
            ranking,
            classifier,
            query_strategy,
            balance_strategy,
            feature_extraction,
            training_set,
        )

        last_ranking = state.get_last_ranking()
        assert isinstance(last_ranking, pd.DataFrame)
        assert list(last_ranking.columns) == [
            "record_id",
            "ranking",
            "classifier",
            "query_strategy",
            "balance_strategy",
            "feature_extraction",
            "training_set",
            "time",
        ]

        assert last_ranking["ranking"].to_list() == [0, 1, 2, 3, 4, 5]
        assert last_ranking["record_id"].to_list() == ranking
        assert last_ranking["classifier"].to_list() == [classifier] * len(record_ids)


def test_get_pool():
    with open_state(TEST_STATE_FP) as state:
        pool = state.get_pool()

    assert isinstance(pool, pd.Series)
    assert len(pool) == 841
    assert pool[:10].to_list() == TEST_POOL_START


def test_get_labeled():
    with open_state(TEST_STATE_FP) as state:
        labeled = state.get_labeled()

    assert isinstance(labeled, pd.DataFrame)
    assert labeled["record_id"].to_list() == TEST_RECORD_IDS
    assert labeled["label"].to_list() == TEST_LABELS


def test_add_extra_column(tmpdir):
    """Check if state still works with extra colums added to tables."""
    project_path = Path(tmpdir, "test.asreview")
    ASReviewProject.create(project_path)

    with open_state(project_path, read_only=False) as state:
        con = state._connect_to_sql()
        cur = con.cursor()
        cur.execute("ALTER TABLE last_ranking ADD COLUMN test_lr INTEGER;")
        cur.execute("ALTER TABLE results ADD COLUMN test_res INTEGER;")
        con.commit()
        con.close()

    record_ids = [1, 2, 3, 4, 5, 6]
    ranking = [1, 3, 4, 6, 2, 5]
    classifier = "nb"
    query_strategy = "max"
    balance_strategy = "double"
    feature_extraction = "tfidf"
    training_set = 2

    with open_state(project_path, read_only=False) as state:
        state.add_record_table(record_ids)
        state.add_last_ranking(
            ranking,
            classifier,
            query_strategy,
            balance_strategy,
            feature_extraction,
            training_set,
        )

        top_ranked = state.query_top_ranked(1)
        pool, labeled, pending = state.get_pool_labeled_pending()
        assert len(pending) == 1
        assert len(pool) == len(record_ids) - 1
        assert len(labeled) == 0

        state.add_labeling_data(top_ranked, [0 for _ in top_ranked])
        pool, labeled, pending = state.get_pool_labeled_pending()
        assert len(pending) == 0
        assert len(pool) == len(record_ids) - 1
        assert len(labeled) == 1
